from bs4 import BeautifulSoup

html_file = "./build/coverage/index.html"

colour_dict = {'coverage-low':"red", 'coverage-medium':"yellow", 'coverage-high':"brightgreen"}

with open(html_file, 'r') as file:
    soup = BeautifulSoup(file, 'html.parser')

    # Find the "coverage" table
    coverage_table = soup.find('table', {'class': 'coverage'})

    # Find the second row of the "coverage" table
    second_row = coverage_table.find_all('tr')[1]

    # Find the third column of the second row
    third_column = second_row.find_all('td')[2]

    percentage = third_column.get_text()
    
    # Get the class of the <td> element
    td_class = third_column.get('class')[0]

print(int(float(percentage.replace("%", ""))))


# %%
import numpy as np
from scipy.interpolate import griddata
from scipy.spatial import cKDTree
from mayavi import mlab
# from mayavi.api import Engine



# %%
a = 1.0
b = 1.0
c = 1.0
n_x = 0
n_y = 0
n_z = 0

# Read the data from the file
data = []
atoms = []
species_list = []
species_name_list = []
with open('../viability_BTO.dat', 'r') as file:
    for line in file:
        # if line starts with #grid, read the three values as n_x, n_y, n_z
        if line.startswith("#grid"):
            values = line.strip().split()
            n_x = int(values[1])
            n_y = int(values[2])
            n_z = int(values[3])
        # if line starts with #lat, read the three values as a, b, c
        if line.startswith("#lat"):
            values = line.strip().split()
            a = float(values[1])
            b = float(values[2])
            c = float(values[3])
        if line.startswith("#species"):
            values = line.strip().split()
            species_name_list = values[1:]
        if line.startswith("#"):
            continue
        values = line.strip().split()
        if len(values) == 3:
            atoms.append([float(values[0]), float(values[1]), float(values[2])])
        if len(values) >= 4:
            data.append([float(values[0]), float(values[1]), float(values[2]), float(values[3])])
            if len(values) > 4:
                species_list.append(int(values[4]))

# scale atom locations
atoms = np.array(atoms)
atoms[:,0] = atoms[:,0] * n_x
atoms[:,1] = atoms[:,1] * n_y
atoms[:,2] = atoms[:,2] * n_z

# set the min and max values for the grid
x_min = a * ( 0.0 )
x_max = a * ( 1.0 - 1.0/n_x )
y_min = b * ( 0.0 )
y_max = b * ( 1.0 - 1.0/n_y )
z_min = c * ( 0.0 )
z_max = c * ( 1.0 - 1.0/n_z )

# Create a 3D grid
grid_x, grid_y, grid_z = np.mgrid[x_min:x_max:complex(n_x), y_min:y_max:complex(n_y), z_min:z_max:complex(n_z)]
grid_points = np.vstack((grid_x.ravel(), grid_y.ravel(), grid_z.ravel())).T


# %%
# Function to plot atoms as spheres
def plot_atoms_as_spheres(ax, atoms, radius=0.5, resolution=(10, 5)):
    u = np.linspace(0, 2 * np.pi, resolution[0])
    v = np.linspace(0, np.pi, resolution[1])
    for atom in atoms:
        atom_x, atom_y, atom_z = atom
        # Generate sphere coordinates
        x_sphere = atom_x + radius * np.outer(np.cos(u), np.sin(v))
        y_sphere = atom_y + radius * np.outer(np.sin(u), np.sin(v))
        z_sphere = atom_z + radius * np.outer(np.ones(np.size(u)), np.cos(v))
        # Plot spheres with lower resolution
        ax.plot_surface(x_sphere, y_sphere, z_sphere, color='red', alpha=1.0, rstride=1, cstride=1)
        
def plot_atoms_as_scatter(ax, atoms, size=100, color='red'):
    ax.scatter(atoms[:,0], atoms[:,1], atoms[:,2], s=size, c=color, alpha=1.0)

# Get atom bonds within radius
def get_bonds(atoms, grid, abc, radius=1.0):
    bonds = []
    for i, atom1 in enumerate(atoms):
        for j, atom2 in enumerate(atoms):
            if i != j:
                distance = np.linalg.norm( ( atom1 - atom2 ) * abc / grid)
                if distance <= radius:
                    bonds.append((i, j))
    return bonds

# Plot the atoms as spheres
def plot_atoms(atoms, bonds, radius=0.1, color=(1, 0, 0)):
    x_list = []
    y_list = []
    z_list = []
    for atom in atoms:
        x_sphere, y_sphere, z_sphere = atom
        x_list.append(x_sphere)
        y_list.append(y_sphere)
        z_list.append(z_sphere)
        # mlab.points3d(x_sphere, y_sphere, z_sphere, scale_factor=radius, color=color)
    pts = mlab.points3d(x_list, y_list, z_list, scale_factor=radius, color=color)

    pts.mlab_source.dataset.lines = np.array(bonds)

    tube = mlab.pipeline.tube(pts, tube_radius=1.0)
    tube.filter.radius_factor = 1.
    mlab.pipeline.surface(tube, color=color)


# # Draw the lattice bounding box
def draw_bounding_box(a, b, c):
    # Define the vertices of the bounding box
    vertices = np.array([[0.0, 0.0, 0.0], [a, 0.0, 0.1], [a, b, 0.0], [0.0, b, 0.0],
                         [0.0, 0.0, c], [a, 0.0, c], [a, b, c], [0.0, b, c]])
    
    # Draw the lines connecting the vertices to create the box
    edges = [
        [vertices[0], vertices[1]], [vertices[1], vertices[2]], [vertices[2], vertices[3]], [vertices[3], vertices[0]],  # Bottom face
        [vertices[4], vertices[5]], [vertices[5], vertices[6]], [vertices[6], vertices[7]], [vertices[7], vertices[4]],  # Top face
        [vertices[0], vertices[4]], [vertices[1], vertices[5]], [vertices[2], vertices[6]], [vertices[3], vertices[7]],  # Vertical edges
    ]
    
    connections = ((0, 1), (1, 2), (2, 3), (3, 0), (0, 4), (1, 5), (2, 6), (3, 7), (4, 5), (5, 6), (6, 7), (7, 4), (4, 0), (5, 1), (6, 2), (7, 3))
    x_list = []
    y_list = []
    z_list = []
    for vertex in vertices:
        x_list.append(vertex[0])
        y_list.append(vertex[1])
        z_list.append(vertex[2])
    pts = mlab.points3d(x_list, y_list, z_list, scale_factor=0.1, color=(0, 0, 0))
    pts.mlab_source.dataset.lines = np.array(connections)
    tube = mlab.pipeline.tube(pts, tube_radius=0.5)
    tube.filter.radius_factor = 1.
    mlab.pipeline.surface(tube, color=(0, 0, 0))

    # for edge in edges:
    #     mlab.plot3d(*zip(*edge), tube_radius=0.01, color=(0, 0, 0))

# %%

grid_values = []
distance_threshold = 1e-3
for spec in sorted(set(species_list)):
    print("Plotting species: ", spec)
    # Extract the coordinates and values
    x = np.array([row[0] for row, species in zip(data, species_list) if species == spec])
    y = np.array([row[1] for row, species in zip(data, species_list) if species == spec])
    z = np.array([row[2] for row, species in zip(data, species_list) if species == spec])
    values = np.array([row[3] for row, species in zip(data, species_list) if species == spec])

    # scale the data positions
    x = x * a
    y = y * b
    z = z * c

    # Calculate distances to the nearest known data point
    tree = cKDTree(np.c_[x, y, z])
    distances, _ = tree.query(grid_points, k=1)

    # Interpolate data onto the 3D grid
    grid_values.append(griddata((x, y, z), values, (grid_x, grid_y, grid_z), method='nearest', fill_value=0.0))
    # grid_values[spec] = np.nan_to_num(grid_values)

    # Reshape distances to match the grid shape
    distances = distances.reshape(grid_values[spec-1].shape)

    # Set threshold for distance (e.g., 1 unit)
    grid_values[spec-1][distances > distance_threshold] = 0  # Set to zero for points beyond the threshold

    mlab.contour3d(grid_values[spec-1], contours=10, transparent=True)
    # ax1 = mlab.axes( color=(1,1,1), nb_labels=4 )

# Call the function to plot atoms
bonds = get_bonds(atoms, [n_x, n_y, n_z], [a, b, c], radius=2.5)
plot_atoms(atoms, bonds, radius=10.0)

# Call the function to draw the bounding box
draw_bounding_box(n_x, n_y, n_z)
# draw_bounding_box()

mlab.show()



from chgnet.model import CHGNetCalculator
from ase.calculators.singlepoint import SinglePointCalculator
from raffle.generator import raffle_generator
from ase import build, Atoms
from ase.optimize import FIRE
from ase.io import write, read
import numpy as np
import os
from joblib import Parallel, delayed


# Function to relax a structure
def process_structure(i, atoms, num_structures_old, calc_params, optimise_structure, iteration):
    # Check if the structure has already been processed
    if i < num_structures_old:
        return None, None, None
    
    # calc = Vasp(**calc_params, label=f"struct{i}", directory=f"iteration{iteration}/struct{i}/", txt=f"stdout{i}.o")
    inew = i - num_structures_old
    atoms.calc = calc

    # Calculate and save the initial energy per atom
    energy_unrlxd = atoms.get_potential_energy() / len(atoms)
    print(f"Initial energy per atom: {energy_unrlxd}")

    # Optimise the structure if requested
    if optimise_structure:
        optimizer = FIRE(atoms, trajectory = f"traje{inew}.traj", logfile=f"optimisation{inew}.log")
        try:
            optimizer.run(fmax=0.05, steps=100)
        except Exception as e:
            print(f"Optimisation failed: {e}")
            return None, None, None
    
    # Save the optimised structure and its energy per atom
    energy_rlxd = atoms.get_potential_energy() / len(atoms)

    # Get the distance matrix
    distances = atoms.get_all_distances(mic=True)

    # Set self-distances (diagonal entries) to infinity to ignore them
    np.fill_diagonal(distances, np.inf)

    # Check if the minimum non-self distance is below 1.5
    if distances.min() < 1.0:
        print(f"Distance too small: {atoms.get_all_distances(mic=True).min()}")
        return None, None, None
    
    if abs(energy_rlxd - energy_unrlxd) > 10.0:
        print(f"Energy difference too large: {energy_rlxd} vs {energy_unrlxd}")
        return None, None, None
    
    return atoms, energy_unrlxd, energy_rlxd


if __name__ == "__main__":

    # set up the calculator
    calc_params = {}
    calc = CHGNetCalculator()

    hosts = []
    # get all hosts in ../../data/C-MgO_hosts/ directory
    for filename in os.listdir("../../data/C-MgO_hosts/"):
        if filename.startswith("POSCAR"):
            atoms = read(f"../../data/C-MgO_hosts/{filename}")
            if not all(element == 'C' for element in set(atoms.get_chemical_symbols())):
                print(f"Skipping {filename} as it has non-carbon atoms")
                continue
            elif len(atoms) > 20:
                print(f"Skipping {filename} as it has more than 20 atoms (want to do smaller cells for now)")
                continue
            hosts.append(read(f"../../data/C-MgO_hosts/{filename}"))
            hosts[-1].calc = calc

    # set up element reference energies
    C_reference = Atoms('C4', positions=[
        [0.000000,    0.000000,    1.950768],
        [0.000000,    0.000000,    5.852305],
        [0.000000,    1.424491,    1.950768],
        [1.233646,    0.712246,    5.852305],
        ], cell=[[1.2336, -2.1367, 0.0], [1.2336, 2.1367, 0.0], [0.0, 0.0, 7.8031]], pbc=True)
    C_reference.calc = calc
    C_reference_energy = C_reference.get_potential_energy() / len(C_reference)
    Mg_reference = Atoms("Mg2", positions=[
            [0.000000,    1.831369,    1.285301],
            [1.586012,    0.915684,    3.855904],
        ], cell=[
            [1.58601,   -2.7471,    0.0000],
            [1.58601,    2.7471,    0.0000],
            [0.00000,    0.0001,    5.1412],
        ], pbc=True)
    Mg_reference.calc = calc
    Mg_reference_energy = Mg_reference.get_potential_energy() / len(Mg_reference)
    O_reference = Atoms("O2", positions=[
        [0.0, 0.0, 0.0],
        [1.23, 0.0, 0.0]], cell=[10, 10, 10], pbc=False)
    O_reference.calc = calc
    O_reference_energy = O_reference.get_potential_energy() / len(O_reference)

    generator = raffle_generator()
    generator.distributions.set_element_energies(
        {
            'C': C_reference_energy,
            'Mg': Mg_reference_energy,
            'O': O_reference_energy
        }
    )

    # set the covalent radii
    generator . distributions . set_bond_radii (
        {
            ('C', 'Mg'): 1.4,
            ('Mg', 'Mg'): 1.4,
            ('C', 'O'): 1.4,
            # ('C', 'Mg'): 1.15, # 2.3 = what Joe used https://github.com/ExeQuantCode/RAFFLE/blob/1cfdf2d220dce41f20ace4cce5a4135a37d1f1b1/chem.in
            # ('C', 'O'): 0.587,
            # ('Mg', 'Mg'): 1.592,
            # INITIALISE O2 BOND TO LARGER VALUE
        }
    )

    # set energy scale
    generator.distributions.set_kBT(0.4)

    # set the distribution function widths (2-body, 3-body, 4-body)
    generator.distributions.set_width([0.025, np.pi/200.0, np.pi/200.0])

    # generate MgO bulk
    MgO_atom = Atoms("Mg4O4", positions=[
            [0.0, 0.0, 0.0],
            [0.0, 2.097, 2.097],
            [2.097, 0.0, 2.097],
            [2.097, 2.097, 0.0],
            [0.0, 0.0, 2.097],
            [0.0, 2.097, 0.0],
            [2.097, 0.0, 0.0],
            [2.097, 2.097, 2.097],
        ], cell=[4.1940, 4.1940, 4.1940], pbc=True)
    MgO_atom.calc = calc

    # read C-Mg-O database, obtained from Materials Project
    print("Reading database")
    database = read("../database.xyz", index=":")

    # set up the initial database
    initial_database = []
    for i, atoms in enumerate(database):
        if atoms.calc is None:
            database.remove(atoms)
            continue
        # check if any other species in the system other than C, Mg, O
        if not all(element in ['C', 'Mg', 'O'] for element in set(atoms.get_chemical_symbols())):
            print(f"Skipping structure {i} as it has non-C-Mg-O atoms")
            continue
        atoms.calc = calc
        initial_database.append(atoms)
    initial_database.append(C_reference)
    initial_database.append(Mg_reference)
    initial_database.append(O_reference)
    initial_database.append(MgO_atom)
    generator.distributions.create(initial_database, deallocate_systems=False)

    # set parameters
    seed = 0

    # check if the energies file exists, if not create it
    energies_rlxd_filename = f"energies_rlxd_seed{seed}.txt"
    energies_unrlxd_filename = f"energies_unrlxd_seed{seed}.txt"
    if os.path.exists(energies_rlxd_filename):
        with open(energies_rlxd_filename, "w") as energy_file:
            pass
    else:
        open(energies_rlxd_filename, "w").close()

    if os.path.exists(energies_unrlxd_filename):
        with open(energies_unrlxd_filename, "w") as energy_file:
            pass
    else:
        open(energies_unrlxd_filename, "w").close()
    
    # initialise the number of structures generated
    iter = -1
    num_structures_old = 0
    unrlxd_structures = []
    rlxd_structures = []
    optimise_structure = True
    mass_C = 12.011
    mass_Mg = 24.305
    mass_O = 15.999
    # loop over all hosts
    for host in hosts:
        print("setting host")
        generator.set_host(host)
        volume = host.get_volume()
        num_atoms_C = len(host)
    
        host_positions = host.get_scaled_positions(wrap=True)
        host_height = np.linalg.norm(host.get_cell()[2, :])
        print(f"host_height: {host_height}")

        # get all unique z values
        z_values = np.unique(host_positions[:, 2])
        print(f"z_values: {z_values}")
        # return the two closes to 0.5 in the cell
        z_values = z_values[np.argsort(np.abs(z_values - 0.5))[:2]]
        print(f"z_values: {z_values}")
        # set the bounds for the generator
        z_min = min(z_values) + 1.0 / host_height
        z_max = max(z_values) - 1.0 / host_height
        generator.set_bounds([[0, 0, z_min], [1, 1, z_max]])

        # loop over range of MgO densities to intercalate
        for num_atoms in range(1, 50):

            num_atoms_Mg = num_atoms
            num_atoms_O = num_atoms
            density = ( mass_C * num_atoms_C + mass_Mg * num_atoms_Mg + mass_O * num_atoms_O ) / volume
            if density > 2.2:
                # print("Density too high:", density, "u/A^3")
                continue
            elif density < 1.0:
                # print("Density too low", density, "u/A^3")
                continue
            iter += 1

            # generate structures
            generator.generate(
                num_structures = 5,
                stoichiometry = { 'Mg': num_atoms_Mg, 'O': num_atoms_O },
                seed = seed*1000+iter,
                method_ratio = {"void": 1.0, "rand": 0.001, "walk": 1.0, "grow": 0.0, "min": 1.0},
                verbose = 0,
            )

            # print the number of structures generated
            print("Total number of structures generated: ", generator.num_structures)
            generated_structures = generator.get_structures(calc)
            num_structures_new = len(generated_structures)

            # check if directory iteration[iter] exists, if not create it
            iterdir = f"iteration{iter}/"
            if not os.path.exists(iterdir):
                os.makedirs(iterdir)
            generator.print_settings(iterdir+"generator_settings.txt")

            # set up list of energies
            energy_unrlxd = np.zeros(num_structures_new - num_structures_old)
            energy_rlxd = np.zeros(num_structures_new - num_structures_old)
            for i in range(num_structures_new - num_structures_old):
                write(iterdir+f"POSCAR_unrlxd_{i}", generated_structures[num_structures_old + i])
                print(f"Structure {i} energy per atom: {generated_structures[num_structures_old + i].get_potential_energy() / len(generated_structures[num_structures_old + i])}")
                unrlxd_structures.append(generated_structures[num_structures_old + i].copy())
                unrlxd_structures[-1].calc = SinglePointCalculator(
                    generated_structures[num_structures_old + i],
                    energy=generated_structures[num_structures_old + i].get_potential_energy(),
                    forces=generated_structures[num_structures_old + i].get_forces()
            
            # Start parallel execution
            print("Starting parallel execution")
            results = Parallel(n_jobs=5)(
                delayed(process_structure)(i, generated_structures[i].copy(), num_structures_old, calc_params, optimise_structure, iteration=seed)
                for i in range(num_structures_old, num_structures_new)
            )

            # Wait for all futures to complete
            for j, result in enumerate(results):
                generated_structures[num_structures_old + j], energy_unrlxd[j], energy_rlxd[j] = result
                if generated_structures[num_structures_old + j] is None:
                    print("Structure failed the checks")
                    continue
                rlxd_structures.append(generated_structures[num_structures_old + j].copy())
                rlxd_structures[-1].calc = SinglePointCalculator(
                    generated_structures[j+num_structures_old],
                    energy=generated_structures[num_structures_old + j].get_potential_energy(),
                    forces=generated_structures[num_structures_old + j].get_forces()
                )
            print("All futures completed")

            # Remove structures that failed the checks
            for j, atoms in reversed(list(enumerate(generated_structures))):
                if j < num_structures_old:
                    continue
                if atoms is None:
                    energy_unrlxd = np.delete(energy_unrlxd, j-num_structures_old)
                    energy_rlxd = np.delete(energy_rlxd, j-num_structures_old)
                    del generated_structures[j]
                    del unrlxd_structures[j]
                    del rlxd_structures[j]
                    generator.remove_structure(j)
            num_structures_new = len(generated_structures) 

            # write the structures to files
            for i in range(num_structures_new - num_structures_old):
                write(iterdir+f"POSCAR_{i}", generated_structures[num_structures_old + i])
                print(f"Structure {i} energy per atom: {energy_rlxd[i]}")
                # append energy per atom to the 'energies_unrlxd_filename' file
                with open(energies_unrlxd_filename, "a") as energy_file:
                    energy_file.write(f"{i+num_structures_old} {energy_unrlxd[i]}\n")
                # append energy per atom to the 'energies_rlxd_filename' file
                with open(energies_rlxd_filename, "a") as energy_file:
                    energy_file.write(f"{i+num_structures_old} {energy_rlxd[i]}\n")

            # update the distribution functions
            print("Updating distributions")
            generator.distributions.update(generated_structures[num_structures_old:], from_host=False, deallocate_systems=False)

            # print the new distribution functions to a file
            print("Printing distributions")
            generator.distributions.write_dfs(iterdir+"distributions.txt")
            generator.distributions.write_2body(iterdir+"df2.txt")
            generator.distributions.write_3body(iterdir+"df3.txt")
            generator.distributions.write_4body(iterdir+"df4.txt")
            generator.distributions.deallocate_systems()

            # update the number of structures generated
            num_structures_old = num_structures_new

    # Write the final distribution functions to a file
    generator.distributions.write_gdfs(f"gdfs_seed{seed}.txt")

    # Read energies from the file
    with open(energies_rlxd_filename, "r") as energy_file:
        energies = energy_file.readlines()

    # Parse and sort the energies
    energies = [line.strip().split() for line in energies]
    energies = sorted(energies, key=lambda x: float(x[1]))

    # Write the sorted energies back to the file
    with open(f"sorted_{energies_rlxd_filename}", "w") as energy_file:
        for entry in energies:
            energy_file.write(f"{int(entry[0])} {float(entry[1])}\n")

    # Write the structures to files
    write(f"unrlxd_structures_seed{seed}.traj", unrlxd_structures)
    write(f"rlxd_structures_seed{seed}.traj", rlxd_structures)
    print("All generated and relaxed structures written")

    print("Learning complete")


from __future__ import print_function, absolute_import, division
import raffle._raffle as _raffle
import f90wrap.runtime
import logging
import numpy
import warnings


class Geom_Rw(f90wrap.runtime.FortranModule):
    """
    Code for handling geometry read/write operations.

    This module provides the necessary functionality to read, write, and
    store atomic geometries.
    In this module, and all of the codebase, element and species are used
    interchangeably.

    Defined in ../src/lib/mod_geom_rw.f90

    .. note::
        It is recommended not to use this module directly, but to handle
        atom objects through the ASE interface.
        This is provided mostly for compatibility with the existing codebase
        and Fortran code.
    """
    @f90wrap.runtime.register_class("raffle.species_type")
    class species_type(f90wrap.runtime.FortranDerivedType):
        def __init__(self, handle=None):
            """
            Create a ``species_type`` object.

            Returns:
                species (species_type):
                    Object to be constructed
            """
            f90wrap.runtime.FortranDerivedType.__init__(self)
            result = _raffle.f90wrap_geom_rw__species_type_initialise()
            self._handle = result[0] if isinstance(result, tuple) else result

        def __del__(self):
            """
            Destructor for class species_type


            Defined at ../src/lib/mod_geom_rw.f90 lines \
                26-32

            Parameters
            ----------
            this : species_type
            	Object to be destructed


            Automatically generated destructor for species_type
            """
            if self._alloc:
                _raffle.f90wrap_geom_rw__species_type_finalise(this=self._handle)

        @property
        def atom(self):
            """
            Derived type containing the atomic information of a crystal.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_species_type__array__atom(self._handle)
            if array_handle in self._arrays:
                atom = self._arrays[array_handle]
            else:
                atom = f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_species_type__array__atom)
                self._arrays[array_handle] = atom
            return atom

        @atom.setter
        def atom(self, atom):
            self.atom[...] = atom

        @property
        def mass(self):
            """
            The mass of the element.
            """
            return _raffle.f90wrap_species_type__get__mass(self._handle)

        @mass.setter
        def mass(self, mass):
            _raffle.f90wrap_species_type__set__mass(self._handle, mass)

        @property
        def charge(self):
            """
            The charge of the element.
            """
            return _raffle.f90wrap_species_type__get__charge(self._handle)

        @property
        def radius(self):
            """
            The radius of the element.
            """
            return _raffle.f90wrap_species_type__get__radius(self._handle)

        @radius.setter
        def radius(self, radius):
            _raffle.f90wrap_species_type__set__radius(self._handle, radius)

        @charge.setter
        def charge(self, charge):
            _raffle.f90wrap_species_type__set__charge(self._handle, charge)

        @property
        def name(self):
            """
            The symbol of the element.
            """
            return _raffle.f90wrap_species_type__get__name(self._handle)

        @name.setter
        def name(self, name):
            _raffle.f90wrap_species_type__set__name(self._handle, name)

        @property
        def num(self):
            """
            The number of atoms of this species/element.
            """
            return _raffle.f90wrap_species_type__get__num(self._handle)

        @num.setter
        def num(self, num):
            _raffle.f90wrap_species_type__set__num(self._handle, num)

        def __str__(self):
            ret = ['<species_type>{\n']
            ret.append('    atom : ')
            ret.append(repr(self.atom))
            ret.append(',\n    mass : ')
            ret.append(repr(self.mass))
            ret.append(',\n    charge : ')
            ret.append(repr(self.charge))
            ret.append(',\n    name : ')
            ret.append(repr(self.name))
            ret.append(',\n    num : ')
            ret.append(repr(self.num))
            ret.append('}')
            return ''.join(ret)

        _dt_array_initialisers = []


    @f90wrap.runtime.register_class("raffle.basis")
    class basis(f90wrap.runtime.FortranDerivedType):
        def __init__(self, atoms=None, handle=None):
            """
            Create a ``basis`` object.

            This object is used to store the atomic information of a crystal,
            including lattice and basis information.
            This is confusingly named as a crystal = lattice + basis.

            Returns:
                basis (basis):
                    Object to be constructed
            """
            f90wrap.runtime.FortranDerivedType.__init__(self)
            result = _raffle.f90wrap_geom_rw__basis_type_initialise()
            self._handle = result[0] if isinstance(result, tuple) else result

            if atoms is not None:
                self.fromase(atoms)

        def __del__(self):
            """
            Destructor for class basis


            Defined at ../src/lib/mod_geom_rw.f90 lines \
                34-42

            Parameters
            ----------
            this : basis
            	Object to be destructed


            Automatically generated destructor for basis
            """
            if self._alloc:
                _raffle.f90wrap_geom_rw__basis_type_finalise(this=self._handle)

        def allocate_species(self, num_species=None, species_symbols=None, species_count=None, \
            positions=None):
            """
            Allocate memory for the species list.

            Parameters:
                num_species (int):
                    Number of species
                species_symbols (list of str):
                    List of species symbols
                species_count (list of int):
                    List of species counts
                atoms (list of float):
                    List of atomic positions
            """
            _raffle.f90wrap_geom_rw__allocate_species__binding__basis_type(this=self._handle, \
                num_species=num_species, species_symbols=species_symbols, species_count=species_count, \
                atoms=positions)

        def _init_array_spec(self):
            """
            Initialise the species array.
            """
            self.spec = f90wrap.runtime.FortranDerivedTypeArray(self,
                                            _raffle.f90wrap_basis_type__array_getitem__spec,
                                            _raffle.f90wrap_basis_type__array_setitem__spec,
                                            _raffle.f90wrap_basis_type__array_len__spec,
                                            """
            Element spec ftype=type(species_type) pytype=species_type


            Defined at ../src/lib/mod_geom_rw.f90 line 35

            """, Geom_Rw.species_type)
            return self.spec

        def toase(self, calculator=None):
            """
            Convert the basis object to an ASE Atoms object.

            Parameters:
                calculator (ASE Calculator):
                    ASE calculator object to be assigned to the Atoms object.
            """
            from ase import Atoms

            # Set the species list
            positions = []
            species_string = ""
            for i in range(self.nspec):
                for j in range(self.spec[i].num):
                    species_string += str(self.spec[i].name.decode()).strip()
                    positions.append(self.spec[i].atom[j])

            # Set the atoms
            if(self.lcart):
                atoms = Atoms(species_string, positions=positions, cell=self.lat, pbc=self.pbc)
            else:
                atoms = Atoms(species_string, scaled_positions=positions, cell=self.lat, pbc=self.pbc)

            if calculator is not None:
                atoms.calc = calculator
            return atoms

        def fromase(self, atoms, verbose=False):
            """
            Convert the ASE Atoms object to a basis object.

            Parameters:
                atoms (ASE Atoms):
                    ASE Atoms object to be converted.
                verbose (bool):
                    Boolean whether to print warnings.
            """
            from ase.calculators.singlepoint import SinglePointCalculator

            # Get the species symbols
            species_symbols = atoms.get_chemical_symbols()
            species_symbols_unique = sorted(set(species_symbols))

            # Set the number of species
            self.nspec = len(species_symbols_unique)

            # Set the number of atoms
            self.natom = len(atoms)

            # check if calculator is present
            if atoms.calc is None:
                if verbose:
                    print("WARNING: No calculator present, setting energy to 0.0")
                atoms.calc = SinglePointCalculator(atoms, energy=0.0)
            self.energy = atoms.get_potential_energy()

            # # Set the lattice vectors
            self.lat = numpy.reshape(atoms.get_cell().flatten(), [3,3], order='A')
            self.pbc = atoms.pbc

            # Set the system name
            self.sysname = atoms.get_chemical_formula()

            # Set the species list
            species_count = []
            atom_positions = []
            positions = atoms.get_scaled_positions()
            for species in species_symbols_unique:
                species_count.append(sum([1 for symbol in species_symbols if symbol == species]))
                for j, symbol in enumerate(species_symbols):
                    if symbol == species:
                        atom_positions.append(positions[j])

            # Allocate memory for the atom list
            self.lcart = False
            self.allocate_species(species_symbols=species_symbols_unique, species_count=species_count, positions=atom_positions)

        @property
        def nspec(self):
            """
            The number of species in the basis.
            """
            return _raffle.f90wrap_basis_type__get__nspec(self._handle)

        @nspec.setter
        def nspec(self, nspec):
            _raffle.f90wrap_basis_type__set__nspec(self._handle, nspec)

        @property
        def natom(self):
            """
            The number of atoms in the basis.
            """
            return _raffle.f90wrap_basis_type__get__natom(self._handle)

        @natom.setter
        def natom(self, natom):
            _raffle.f90wrap_basis_type__set__natom(self._handle, natom)

        @property
        def energy(self):
            """
            The energy associated with the basis (or crystal).
            """
            return _raffle.f90wrap_basis_type__get__energy(self._handle)

        @energy.setter
        def energy(self, energy):
            _raffle.f90wrap_basis_type__set__energy(self._handle, energy)

        @property
        def lat(self):
            """
            The lattice vectors of the basis.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_basis_type__array__lat(self._handle)
            if array_handle in self._arrays:
                lat = self._arrays[array_handle]
            else:
                lat = f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_basis_type__array__lat)
                self._arrays[array_handle] = lat
            return lat

        @lat.setter
        def lat(self, lat):
            self.lat[...] = lat

        @property
        def lcart(self):
            """
            Boolean whether the atomic positions are in cartesian coordinates.
            """
            return _raffle.f90wrap_basis_type__get__lcart(self._handle)

        @lcart.setter
        def lcart(self, lcart):
            _raffle.f90wrap_basis_type__set__lcart(self._handle, lcart)

        @property
        def pbc(self):
            """
            Boolean array indicating the periodic boundary conditions.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_basis_type__array__pbc(self._handle)
            if array_handle in self._arrays:
                pbc = self._arrays[array_handle]
            else:
                pbc = f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_basis_type__array__pbc)
                self._arrays[array_handle] = pbc
            return pbc

        @pbc.setter
        def pbc(self, pbc):
            self.pbc[...] = pbc

        @property
        def sysname(self):
            """
            The name of the system.
            """
            return _raffle.f90wrap_basis_type__get__sysname(self._handle)

        @sysname.setter
        def sysname(self, sysname):
            _raffle.f90wrap_basis_type__set__sysname(self._handle, sysname)

        def __str__(self):
            ret = ['<basis>{\n']
            ret.append('    nspec : ')
            ret.append(repr(self.nspec))
            ret.append(',\n    natom : ')
            ret.append(repr(self.natom))
            ret.append(',\n    energy : ')
            ret.append(repr(self.energy))
            ret.append(',\n    lat : ')
            ret.append(repr(self.lat))
            ret.append(',\n    lcart : ')
            ret.append(repr(self.lcart))
            ret.append(',\n    pbc : ')
            ret.append(repr(self.pbc))
            ret.append(',\n    sysname : ')
            ret.append(repr(self.sysname))
            ret.append('}')
            return ''.join(ret)

        _dt_array_initialisers = [_init_array_spec]



    @f90wrap.runtime.register_class("raffle.basis_array")
    class basis_array(f90wrap.runtime.FortranDerivedType):
        def __init__(self, atoms=None, handle=None):
            """
            Create a ``basis_array`` object.


            Returns:
                basis_array (basis_array):
                    Object to be constructed
            """

            f90wrap.runtime.FortranDerivedType.__init__(self)
            result = _raffle.f90wrap_geom_rw__basis_type_xnum_array_initialise()
            self._handle = result[0] if isinstance(result, tuple) else result


            # check if atoms is an ASE Atoms object or a list of ASE Atoms objects
            if atoms:
                from ase import Atoms
                if isinstance(atoms, Atoms):
                    self.allocate(1)
                    self.items[0].fromase(atoms)
                elif isinstance(atoms, list):
                    self.allocate(len(atoms))
                    for i, atom in enumerate(atoms):
                        self.items[i].fromase(atom)

        def __del__(self):
            """
            Destructor for class basis_array


            Defined at ../src/lib/mod_generator.f90 lines \
                19-21

            Parameters
            ----------
            this : basis_array
            	Object to be destructed


            Automatically generated destructor for basis_array
            """
            if self._alloc:
                _raffle.f90wrap_geom_rw__basis_type_xnum_array_finalise(this=self._handle)

        def _init_array_items(self):
            """
            Initialise the items array.
            """
            self.items = f90wrap.runtime.FortranDerivedTypeArray(self,
                                            _raffle.f90wrap_basis_type_xnum_array__array_getitem__items,
                                            _raffle.f90wrap_basis_type_xnum_array__array_setitem__items,
                                            _raffle.f90wrap_basis_type_xnum_array__array_len__items,
                                            """
            Element items ftype=type(basis_type) pytype=basis


            Defined at  line 0

            """, Geom_Rw.basis)
            return self.items

        def toase(self):
            """
            Convert the basis_array object to a list of ASE Atoms objects.
            """

            # Set the species list
            atoms = []
            for i in range(len(self.items)):
                atoms.append(self.items[i].toase())
            return atoms

        def allocate(self, size):
            """
            Allocate the items array with the given size.

            Parameters:
                size (int):
                    Size of the items array
            """
            _raffle.f90wrap_basis_type_xnum_array__array_alloc__items(self._handle, num=size)

        def deallocate(self):
            """
            Deallocate the items array
            """
            _raffle.f90wrap_basis_type_xnum_array__array_dealloc__items(self._handle)

        _dt_array_initialisers = [_init_array_items]

    _dt_array_initialisers = []


geom_rw = Geom_Rw()

class Raffle__Distribs_Container(f90wrap.runtime.FortranModule):
    """
    Code for handling distribution functions.

    This module provides the necessary functionality to create, update, and
    store distribution functions.
    The distribution functions are used as descriptors for the atomic
    environments in a crystal.
    The generalised distribution function (GDF) is a generalised descriptor
    for the atomic configurations that each species can adopt.

    Defined in ../src/fortran/lib/mod_distribs_container.f90

    """
    @f90wrap.runtime.register_class("raffle.distribs_container_type")
    class distribs_container_type(f90wrap.runtime.FortranDerivedType):
        def __init__(self, handle=None):
            """
            Create a ``Distribs_Container_Type`` object.

            Returns:
                distribution_container (Distribs_Container_Type):
                    Object to be constructed

            """
            f90wrap.runtime.FortranDerivedType.__init__(self)
            result = \
                _raffle.f90wrap_raffle__dc__dc_type_initialise()
            self._handle = result[0] if isinstance(result, tuple) else result

        def __del__(self):
            """
            Destructor for class Distribs_Container_Type


            Defined at ../fortran/lib/mod_distribs_container.f90 \
                lines 25-162

            Parameters
            ----------
            this : Distribs_Container_Type
            	Object to be destructed


            Automatically generated destructor for distribs_container_type
            """
            if self._alloc:
                _raffle.f90wrap_raffle__dc__dc_type_finalise(this=self._handle)

        def set_kBT(self, kBT):
            """
            Set the energy scale for the distribution functions.

            Parameters:
                kBT (float):
                    Energy scale for the distribution functions.

            """
            self.kBT = kBT

        def set_weight_method(self, method):
            """
            Set the weight method for combining the the distribution functions
            to form the generalised distribution function (GDF).

            Parameters:
                method (str):
                    Method to be used for weighting the distribution functions.
                    Allowed values are:
                    - 'formation_energy' or 'formation' or 'form' or 'e_form'
                    - 'energy_above_hull' or 'hull_distance' or 'hull' or 'distance' or 'convex_hull'

            """

            if method in ['empirical', 'formation_energy', 'formation', 'form', 'e_form']:
                self.weight_by_hull = False
            elif method in ['energy_above_hull', 'hull_distance', 'hull', 'distance', 'convex_hull']:
                self.weight_by_hull = True
            else:
                raise ValueError("Invalid weight method: {}".format(method))

        def set_width(self, width):
            """
            Set the distribution function widths.

            Parameters:
                width (list of float):
                    List of distribution function widths.
                    The first element is the 2-body distribution function width,
                    the second element is the 3-body distribution function width,
                    and the third element is the 4-body distribution function width.
            """
            _raffle.f90wrap_raffle__dc__set_width__binding__dc_type(this=self._handle, \
                width=width)

        def set_sigma(self, sigma):
            """
            Set the sigma values of the Gaussians used to
            build the distribution functions.

            Parameters:
                sigma (list of float):
                    List of sigma values.
                    The first element is the 2-body distribution function sigma,
                    the second element is the 3-body distribution function sigma,
                    and the third element is the 4-body distribution function sigma.
            """
            _raffle.f90wrap_raffle__dc__set_sigma__binding__dc_type(this=self._handle, \
                sigma=sigma)

        def set_cutoff_min(self, cutoff_min):
            """
            Set the minimum cutoff values for the distribution functions.

            Parameters:
                cutoff_min (list of float):
                    List of minimum cutoff values.
                    The first element is the 2-body distribution function minimum cutoff,
                    the second element is the 3-body distribution function minimum cutoff,
                    and the third element is the 4-body distribution function minimum cutoff.
            """
            _raffle.f90wrap_raffle__dc__set_cutoff_min__binding__dc_type(this=self._handle, \
                cutoff_min=cutoff_min)

        def set_cutoff_max(self, cutoff_max):
            """
            Set the maximum cutoff values for the distribution functions.

            Parameters:
                cutoff_min (list of float):
                    List of maximum cutoff values.
                    The first element is the 2-body distribution function maximum cutoff,
                    the second element is the 3-body distribution function maximum cutoff,
                    and the third element is the 4-body distribution function maximum cutoff.
            """
            _raffle.f90wrap_raffle__dc__set_cutoff_max__binding__dc_type(this=self._handle, \
                cutoff_max=cutoff_max)

        def set_radius_distance_tol(self, radius_distance_tol):
            """
            Set the radius distance tolerance for the distribution functions.

            The radius distance tolerance represents a multiplier to the bond radii to
            determine the cutoff distance for the distribution functions.

            Parameters:
                radius_distance_tol (list of float):
                    List of radius distance tolerance values.
                    The first two values are the lower and upper bounds for the
                    3-body distribution function radius distance tolerance.
                    The third and fourth values are the lower and upper bounds for the
                    4-body distribution function radius distance tolerance.
            """
            _raffle.f90wrap_raffle__dc__set_radius_distance_tol__binding__dc_type(this=self._handle, \
                radius_distance_tol=radius_distance_tol)

        def create(self, basis_list, energy_above_hull_list=None, deallocate_systems=True, verbose=None):
            """
            Create the distribution functions.

            Parameters:
                basis_list (basis_array or Atoms or list of Atoms):
                    List of atomic configurations to be used to create the distribution functions.
                energy_above_hull_list (list of float):
                    List of energy above hull values for the atomic configurations.
                deallocate_systems (bool):
                    Boolean whether to deallocate the atomic configurations after creating the distribution functions.
                verbose (int):
                    Verbosity level.
            """
            from ase import Atoms
            if isinstance(basis_list, Atoms):
                basis_list = geom_rw.basis_array(basis_list)
            elif isinstance(basis_list, list):
                if all([isinstance(basis, Atoms) for basis in basis_list]):
                    basis_list = geom_rw.basis_array(basis_list)

            _raffle.f90wrap_raffle__dc__create__binding__dc_type(this=self._handle, \
                basis_list=basis_list._handle, \
                energy_above_hull_list=energy_above_hull_list, \
                deallocate_systems=deallocate_systems, \
                verbose=verbose \
            )

        def update(self, basis_list, energy_above_hull_list=None, from_host=True, deallocate_systems=True, verbose=None):
            """
            Update the distribution functions.

            Parameters:
                basis_list (basis_array or Atoms or list of Atoms):
                    List of atomic configurations to be used to create the distribution functions.
                energy_above_hull_list (list of float):
                    List of energy above hull values for the atomic configurations.
                deallocate_systems (bool):
                    Boolean whether to deallocate the atomic configurations after creating the distribution functions.
                from_host (bool):
                    Boolean whether the provided basis_list is based on the host.
                verbose (int):
                    Verbosity level.
            """
            from ase import Atoms
            if isinstance(basis_list, Atoms):
                basis_list = geom_rw.basis_array(basis_list)
            elif isinstance(basis_list, list):
                if all([isinstance(basis, Atoms) for basis in basis_list]):
                    basis_list = geom_rw.basis_array(basis_list)


            _raffle.f90wrap_raffle__dc__update__binding__dc_type(this=self._handle, \
                basis_list=basis_list._handle, \
                energy_above_hull_list=energy_above_hull_list, \
                from_host=from_host, \
                deallocate_systems=deallocate_systems, \
                verbose=verbose \
            )

        def deallocate_systems(self):
            """
            Deallocate the atomic configurations.
            """
            _raffle.f90wrap_raffle__dc__deallocate_systems__binding__dc_type(this=self._handle)

        def add_basis(self, basis):
            """
            Add a basis to the distribution functions.

            It is not recommended to use this function directly, but to use the
            create or update functions instead.

            Parameters:
                basis (geom_rw.basis):
                    Basis object to be added to the distribution functions.

            """
            _raffle.f90wrap_raffle__dc__add_basis__binding__dc_type(this=self._handle, \
                basis=basis._handle)

        def set_element_energies(self, element_energies):
            """
            Set the element reference energies for the distribution functions.

            These energies are used to calculate the formation energies of the
            atomic configurations.

            Parameters:
                element_energies (dict):
                    Dictionary of element reference energies.
                    The keys are the element symbols and the values are the reference energies.
            """

            element_list = list(element_energies.keys())
            energies = [element_energies[element] for element in element_list]
            _raffle.f90wrap_raffle__dc__set_element_energies__binding__dc_type(this=self._handle, \
                elements=element_list, energies=energies)

        def get_element_energies(self):
            """
            Get the element reference energies for the distribution functions.

            Returns:
                element_energies (dict):
                    Dictionary of element reference energies.
                    The keys are the element symbols and the values are the reference energies
            """

            num_elements = _raffle.f90wrap_raffle__dc__get__num_elements(self._handle)
            elements = numpy.zeros((num_elements,), dtype='S3')
            energies = numpy.zeros((num_elements,), dtype=numpy.float32)

            _raffle.f90wrap_raffle__dc__get_element_energies_sm__binding__dc_type(this=self._handle, \
                elements=elements, energies=energies)

            # convert the fortran array to a python dictionary
            element_energies = {}
            for i, element in enumerate(elements):
                name = str(element.decode()).strip()
                element_energies[name] = energies[i]

            return element_energies

        def set_bond_info(self):
            """
            Allocate the bond information array.

            It is not recommended to use this function directly, but to use the
            set_bond_radius or set_bond_radii functions instead.
            """
            _raffle.f90wrap_raffle__dc__set_bond_info__binding__dc_type(this=self._handle)

        def set_bond_radius(self, radius_dict):
            """
            Set the bond radius for the distribution functions.

            Parameters:
                radius_dict (dict):
                    Dictionary of bond radii.
                    The keys are a tuple of the two element symbols and the values are the bond radii.
            """

            # convert radius_dict to elements and radius
            # radius_dict = {('C', 'C'): 1.5}
            elements = list(radius_dict.keys()[0])
            radius = radius_dict.values()[0]

            _raffle.f90wrap_raffle__dc__set_bond_radius__binding__dc_type(this=self._handle, \
                elements=elements, radius=radius)

        def set_bond_radii(self, radius_dict):
            """
            Set the bond radii for the distribution functions.

            Parameters:
                radius_dict (dict):
                    Dictionary of bond radii.
                    The keys are a tuple of the two element symbols and the values are the bond radii.
            """

            # convert radius_list to elements and radii
            # radius_list = {('C', 'C'): 1.5, ('C', 'H'): 1.1}
            elements = []
            radii = []
            for key, value in radius_dict.items():
                elements.append(list(key))
                radii.append(value)


            _raffle.f90wrap_raffle__dc__set_bond_radii__binding__dc_type(this=self._handle, \
                elements=elements, radii=radii)

        def get_bond_radii(self):
            """
            Get the bond radii for the distribution functions.

            Returns:
                bond_radii (dict):
                    Dictionary of bond radii.
                    The keys are a tuple of the two element symbols and the values are the bond radii.
            """

            num_elements = _raffle.f90wrap_raffle__dc__get__num_elements(self._handle)
            if num_elements == 0:
                return {}
            num_pairs = round(num_elements * ( num_elements + 1 ) / 2)
            elements = numpy.zeros((num_pairs,2,), dtype='S3', order='F')
            radii = numpy.zeros((num_pairs,), dtype=numpy.float32, order='F')

            _raffle.f90wrap_raffle__dc__get_bond_radii_staticmem__binding__dc_type(this=self._handle, \
                elements=elements, radii=radii)
            # _raffle.f90wrap_raffle__dc__get_bond_radii_staticmem__binding__dc_type(this=self._handle, \
            #     elements=elements, energies=energies)

            # convert the fortran array to a python dictionary
            bond_radii = {}
            for i, element in enumerate(elements):
                names = tuple([str(name.decode()).strip() for name in element])
                bond_radii[names] = radii[i]

            return bond_radii

        def initialise_gdfs(self):
            """
            Initialise the generalised distribution functions (GDFs).

            It is not recommended to use this function directly, but to use the
            create or update functions instead.
            """
            _raffle.f90wrap_raffle__dc__initialise_gdfs__binding__dc_type(this=self._handle)

        def evolve(self): #, system=None):
            """
            Evolve the distribution functions.

            It is not recommended to use this function directly, but to use the
            create or update functions instead.
            """
            _raffle.f90wrap_raffle__dc__evolve__binding__dc_type(this=self._handle)
            # _raffle.f90wrap_raffle__dc__evolve__binding__dc_type(this=self._handle, \
            #     system=None if system is None else system._handle)

        def write_gdfs(self, file):
            """
            Write the generalised distribution functions (GDFs) to a file.

            Parameters:
                file (str):
                    Name of file to write the GDFs to.
            """
            _raffle.f90wrap_raffle__dc__write_gdfs__binding__dc_type(this=self._handle, \
                file=file)

        def read_gdfs(self, file):
            """
            Read the generalised distribution functions (GDFs) from a file.

            Parameters:
                file (str):
                    Name of file to read the GDFs from.
            """
            _raffle.f90wrap_raffle__dc__read_gdfs__binding__dc_type(this=self._handle, \
                file=file)

        def write_dfs(self, file):
            """
            Write the distribution functions (DFs) associated with all
            allocated systems to a file.

            Parameters:
                file (str):
                    Name of file to write the DFs to.
            """
            _raffle.f90wrap_raffle__dc__write_dfs__binding__dc_type(this=self._handle, \
                file=file)

        def read_dfs(self, file):
            """
            Read the distribution functions (DFs) associated with a set of
            systems from a file.

            Parameters:
                file (str):
                    Name of file to read the DFs from.
            """
            _raffle.f90wrap_raffle__dc__read_dfs__binding__dc_type(this=self._handle, \
                file=file)

        def write_2body(self, file):
            """
            Write the 2-body generalised distribution functions (GDFs) to a file.

            Parameters:
                file (str):
                    Name of file to write the 2-body GDFs to.
            """
            _raffle.f90wrap_raffle__dc__write_2body__binding__dc_type(this=self._handle, \
                file=file)

        def write_3body(self, file):
            """
            Write the 3-body generalised distribution functions (GDFs) to a file.

            Parameters:
                file (str):
                    Name of file to write the 3-body GDFs to.
            """
            _raffle.f90wrap_raffle__dc__write_3body__binding__dc_type(this=self._handle, \
                file=file)

        def write_4body(self, file):
            """
            Write the 4-body generalised distribution functions (GDFs) to a file.

            Parameters:
                file (str):
                    Name of file to write the 4-body GDFs to.
            """
            _raffle.f90wrap_raffle__dc__write_4body__binding__dc_type(this=self._handle, \
                file=file)

        def _get_pair_index(self, species1, species2):
            """
            Get the index of the pair of species in the distribution functions.

            This is meant as an internal function and not likely to be used directly.

            Parameters:
                species1 (str):
                    Name of the first species
                species2 (str):
                    Name of the second species

            Returns:
                idx (int):
                    Index of the pair of species in the distribution functions.
            """
            idx = \
                _raffle.f90wrap_raffle__dc__get_pair_index__binding__dc_type(this=self._handle, \
                species1=species1, species2=species2)
            return idx

        def _get_bin(self, value, dim):
            """
            Get the bin index for a value in the distribution functions.

            This is meant as an internal function and not likely to be used directly.

            Parameters:
                value (float):
                    Value to get the bin index for.
                dim (int):
                    Dimension of the distribution function.
                    1 for 2-body, 2 for 3-body, and 3 for 4-body.

            Returns
            -------
                bin (int):
                    Bin index for the value in the distribution functions.

            """
            bin = \
                _raffle.f90wrap_raffle__dc__get_bin__binding__dc_type(this=self._handle, \
                value=value, dim=dim)
            return bin

        @property
        def num_evaluated(self):
            """
            Number of evaluated distribution functions.
            """
            return _raffle.f90wrap_distribs_container_type__get__num_evaluated(self._handle)

        @num_evaluated.setter
        def num_evaluated(self, num_evaluated):
            _raffle.f90wrap_distribs_container_type__set__num_evaluated(self._handle, \
                num_evaluated)

        @property
        def num_evaluated_allocated(self):
            """
            Number of evaluated distribution functions still allocated.
            """
            return \
                _raffle.f90wrap_distribs_container_type__get__num_evaluated_allocated(self._handle)

        @num_evaluated_allocated.setter
        def num_evaluated_allocated(self, num_evaluated_allocated):
            _raffle.f90wrap_distribs_container_type__set__num_evaluated_allocated(self._handle, \
                num_evaluated_allocated)

        @property
        def kBT(self):
            """
            Energy scale for the distribution functions.
            """
            return _raffle.f90wrap_distribs_container_type__get__kbt(self._handle)

        @kBT.setter
        def kBT(self, kBT):
            _raffle.f90wrap_distribs_container_type__set__kbt(self._handle, kBT)

        @property
        def weight_by_hull(self):
            """
            Boolean whether to weight the distribution functions by the energy above hull.
            """
            return \
                _raffle.f90wrap_distribs_container_type__get__weight_by_hull(self._handle)

        @weight_by_hull.setter
        def weight_by_hull(self, weight_by_hull):
            _raffle.f90wrap_distribs_container_type__set__weight_by_hull(self._handle, \
                weight_by_hull)

        @property
        def nbins(self):
            """
            Number of bins in the distribution functions.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_distribs_container_type__array__nbins(self._handle)
            if array_handle in self._arrays:
                nbins = self._arrays[array_handle]
            else:
                nbins = f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_distribs_container_type__array__nbins)
                self._arrays[array_handle] = nbins
            return nbins

        @nbins.setter
        def nbins(self, nbins):
            self.nbins[...] = nbins

        @property
        def sigma(self):
            """
            Sigma values for the Gaussians used by the
            2-, 3-, and 4-body distribution functions.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_distribs_container_type__array__sigma(self._handle)
            if array_handle in self._arrays:
                sigma = self._arrays[array_handle]
            else:
                sigma = f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_distribs_container_type__array__sigma)
                self._arrays[array_handle] = sigma
            return sigma

        @sigma.setter
        def sigma(self, sigma):
            self.sigma[...] = sigma

        @property
        def width(self):
            """
            Bin widths for the 2-, 3-, and 4-body distribution functions.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_distribs_container_type__array__width(self._handle)
            if array_handle in self._arrays:
                width = self._arrays[array_handle]
            else:
                width = f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_distribs_container_type__array__width)
                self._arrays[array_handle] = width
            return width

        @width.setter
        def width(self, width):
            self.width[...] = width

        @property
        def cutoff_min(self):
            """
            The lower cutoff values for the 2-, 3-, and 4-body distribution functions.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_distribs_container_type__array__cutoff_min(self._handle)
            if array_handle in self._arrays:
                cutoff_min = self._arrays[array_handle]
            else:
                cutoff_min = f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_distribs_container_type__array__cutoff_min)
                self._arrays[array_handle] = cutoff_min
            return cutoff_min

        @cutoff_min.setter
        def cutoff_min(self, cutoff_min):
            self.cutoff_min[...] = cutoff_min

        @property
        def cutoff_max(self):
            """
            The upper cutoff values for the 2-, 3-, and 4-body distribution functions.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_distribs_container_type__array__cutoff_max(self._handle)
            if array_handle in self._arrays:
                cutoff_max = self._arrays[array_handle]
            else:
                cutoff_max = f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_distribs_container_type__array__cutoff_max)
                self._arrays[array_handle] = cutoff_max
            return cutoff_max

        @cutoff_max.setter
        def cutoff_max(self, cutoff_max):
            self.cutoff_max[...] = cutoff_max

        @property
        def radius_distance_tol(self):
            """
            The radius distance tolerance for the distribution functions.

            The radius distance tolerance represents a multiplier to the bond radii to
            determine the cutoff distance for the distribution functions.

            The first two values are the lower and upper bounds for the
            3-body distribution function radius distance tolerance.
            The third and fourth values are the lower and upper bounds for the
            4-body distribution function radius distance tolerance.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_distribs_container_type__array__radius_distance_tol(self._handle)
            if array_handle in self._arrays:
                radius_distance_tol = self._arrays[array_handle]
            else:
                radius_distance_tol = \
                    f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_distribs_container_type__array__radius_distance_tol)
                self._arrays[array_handle] = radius_distance_tol
            return radius_distance_tol

        @radius_distance_tol.setter
        def radius_distance_tol(self, radius_distance_tol):
            self.radius_distance_tol[...] = radius_distance_tol

        def __str__(self):
            ret = ['<distribs_container_type>{\n']
            ret.append('    num_evaluated : ')
            ret.append(repr(self.num_evaluated))
            ret.append(',\n    num_evaluated_allocated : ')
            ret.append(repr(self.num_evaluated_allocated))
            ret.append(',\n    kBT : ')
            ret.append(repr(self.kBT))
            ret.append(',\n    weight_by_hull : ')
            ret.append(repr(self.weight_by_hull))
            ret.append(',\n    nbins : ')
            ret.append(repr(self.nbins))
            ret.append(',\n    sigma : ')
            ret.append(repr(self.sigma))
            ret.append(',\n    width : ')
            ret.append(repr(self.width))
            ret.append(',\n    cutoff_min : ')
            ret.append(repr(self.cutoff_min))
            ret.append(',\n    cutoff_max : ')
            ret.append(repr(self.cutoff_max))
            ret.append(',\n    radius_distance_tol : ')
            ret.append(repr(self.radius_distance_tol))
            ret.append('}')
            return ''.join(ret)

        _dt_array_initialisers = []#_init_array_system]


    _dt_array_initialisers = []


raffle__distribs_container = Raffle__Distribs_Container()

class Generator(f90wrap.runtime.FortranModule):
    """
    Code for generating interface structures.

    The module handles converting Python objects to Fortran objects and vice versa.
    These include converting between dictionaries and Fortran derived types, and
    between ASE Atoms objects and Fortran derived types.

    Defined in ../src/lib/mod_generator.f90

    """
    @f90wrap.runtime.register_class("raffle.stoichiometry_type")
    class stoichiometry_type(f90wrap.runtime.FortranDerivedType):
        def __init__(self, dict=None, element=None, num=None, handle=None):
            """
            Object to store the stoichiometry of a structure.

            Returns:
                stoichiometry (stoichiometry_type):
                    Stoichiometry object
            """
            f90wrap.runtime.FortranDerivedType.__init__(self)
            result = _raffle.f90wrap_stoichiometry_type_initialise()
            self._handle = result[0] if isinstance(result, tuple) else result

            if element:
                self.element = element
            if num:
                self.num = num


        def __del__(self):
            """
            Destructor for class Stoichiometry_Type


            Defined at ../src/lib/mod_generator.f90 lines \
                19-21

            Automatically generated destructor for stoichiometry_type
            """
            if self._alloc:
                _raffle.f90wrap_stoichiometry_type_finalise(this=self._handle)

        @property
        def element(self):
            """
            String representing an element symbol.
            """
            return _raffle.f90wrap_stoichiometry_type__get__element(self._handle)

        @element.setter
        def element(self, element):
            _raffle.f90wrap_stoichiometry_type__set__element(self._handle, element)

        @property
        def num(self):
            """
            Integer representing the number of atoms of the element.
            """
            return _raffle.f90wrap_stoichiometry_type__get__num(self._handle)

        @num.setter
        def num(self, num):
            _raffle.f90wrap_stoichiometry_type__set__num(self._handle, num)

        def __str__(self):
            ret = ['<stoichiometry_type>{\n']
            ret.append('    element : ')
            ret.append(repr(self.element))
            ret.append(',\n    num : ')
            ret.append(repr(self.num))
            ret.append('}')
            return ''.join(ret)

        _dt_array_initialisers = []


    @f90wrap.runtime.register_class("raffle.stoichiometry_array")
    class stoichiometry_array(f90wrap.runtime.FortranDerivedType):
        def __init__(self, dict=None, handle=None):
            """
            Array or list of stoichiometry objects.

            Returns:
                stoichiometry_array (stoichiometry_array):
                    Stoichiometry array object
            """

            f90wrap.runtime.FortranDerivedType.__init__(self)
            result = _raffle.f90wrap_generator__stoich_type_xnum_array_initialise()
            self._handle = result[0] if isinstance(result, tuple) else result
            if dict:
                num_elements = len(dict)
                elements = list(dict.keys())
                nums = list(dict.values())
                self.allocate(num_elements)
                for i in range(num_elements):
                    self.items[i].element = elements[i]
                    self.items[i].num = nums[i]

        def __del__(self):
            """
            Destructor for class Stoichiometry_Type


            Defined at ../src/lib/mod_generator.f90 lines \
                19-21

            Parameters
            ----------
            this : Stoichiometry_Type
            	Object to be destructed


            Automatically generated destructor for stoichiometry_type
            """
            if self._alloc:
                _raffle.f90wrap_generator__stoich_type_xnum_array_finalise(this=self._handle)

        def _init_array_items(self):
            self.items = f90wrap.runtime.FortranDerivedTypeArray(self,
                                            _raffle.f90wrap_stoich_type_xnum_array__array_getitem__items,
                                            _raffle.f90wrap_stoich_type_xnum_array__array_setitem__items,
                                            _raffle.f90wrap_stoich_type_xnum_array__array_len__items,
                                            """
            Element items ftype=type(stoichiometry_type) pytype=stoichiometry_type


            Defined at  line 0

            """, Generator.stoichiometry_type)
            return self.items

        def allocate(self, size):
            """
            Allocate the items array with the given size.

            Parameters:
                size (int):
                    Size of the items array
            """
            _raffle.f90wrap_stoich_type_xnum_array__array_alloc__items(self._handle, num=size)

        def deallocate(self):
            """
            Deallocate the items array.
            """
            _raffle.f90wrap_stoich_type_xnum_array__array_dealloc__items(self._handle)



        _dt_array_initialisers = [_init_array_items]


    @f90wrap.runtime.register_class("raffle.raffle_generator")
    class raffle_generator(f90wrap.runtime.FortranDerivedType):

        def __init__(self, handle=None):
            """
            Create a ``raffle_generator`` object.

            This object is used to generate structures using the RAFFLE method.
            The object has procedures to set the parameters for the generation
            and to generate the structures.
            """
            f90wrap.runtime.FortranDerivedType.__init__(self)
            result = _raffle.f90wrap_generator__raffle_generator_type_initialise()
            self._handle = result[0] if isinstance(result, tuple) else result

        def __del__(self):
            """
            Destructor for class raffle_generator


            Defined at ../src/lib/mod_generator.f90 lines \
                23-34

            """
            if self._alloc:
                _raffle.f90wrap_generator__raffle_generator_type_finalise(this=self._handle)

        def set_max_attempts(self, max_attempts):
            """
            Set the walk-method maximum attempts parameter.
            This parameter determines the maximum number of attempts to generate a structure
            using the walk method before reverting to the void method.

            Parameters:
                max_attempts (int):
                    The maximum number of attempts to generate a structure using the walk method.
            """
            self.max_attempts = max_attempts

        def set_walk_step_size(self, coarse=None, fine=None):
            """
            Set the walk-method step size parameters.

            Parameters:
                coarse (float):
                    The coarse step size for the walk method.
                fine (float):
                    The fine step size for the walk method.
            """
            if coarse is not None:
                self.walk_step_size_coarse = coarse
            if fine is not None:
                self.walk_step_size_fine = fine

        def set_host(self, host):
            """
            Set the host structure for the generation.

            Parameters:
                host (ase.Atoms or geom_rw.basis):
                    The host structure for the generation.
            """
            from ase import Atoms
            # check if host is ase.Atoms object or a Fortran derived type basis_type
            if isinstance(host, Atoms):
                host = geom_rw.basis(atoms=host)

            _raffle.f90wrap_generator__set_host__binding__rgt(this=self._handle, \
                host=host._handle)

        def set_grid(self, grid=None, grid_spacing=None, grid_offset=None):
            """
            Set the grid parameters for the generation.

            Parameters:
                grid (list of int):
                    The number of grid points along each axis of the host.
                grid_spacing (float)
                    The spacing between grid points.
                grid_offset (list of float):
                    The offset of the grid from the origin.
            """
            _raffle.f90wrap_generator__set_grid__binding__raffle_generator_type(this=self._handle, \
                grid=grid, grid_spacing=grid_spacing, grid_offset=grid_offset)

        def reset_grid(self):
            """
            Reset the grid parameters to their default values.
            """
            _raffle.f90wrap_generator__reset_grid__binding__raffle_generator_type(this=self._handle)

        def set_bounds(self, bounds=None):
            """
            Set the bounding box for the generation.

            Parameters:
                bounds (list of list of float):
                    The bounding box within which to constrain placement of atoms.
                    In the form [[a_min, a_max], [b_min, b_max], [c_min, c_max]].
                    Values given in direct (crystal) coordinates, ranging from 0 to 1.
            """
            _raffle.f90wrap_generator__set_bounds__binding__rgt(this=self._handle, \
                bounds=bounds)

        def reset_bounds(self):
            """
            Reset the bounding box to the full host structure.
            """
            _raffle.f90wrap_generator__reset_bounds__binding__rgt(this=self._handle)

        def generate(
                self, num_structures, stoichiometry,
                method_ratio={"void": 0.0, "rand": 0.0, "walk": 0.0, "grow": 0.0, "min": 0.0},
                method_probab=None,
                seed=None, settings_out_file=None, verbose=0,
                calc=None
        ):
            """
            Generate structures using the RAFFLE method.

            Parameters:
                num_structures (int):
                    The number of structures to generate.
                stoichiometry (stoichiometry_array or dict):
                    The stoichiometry of the structures to generate.
                method_ratio (dict):
                    The ratio of using each method to generate a structure.
                method_probab (dict):
                    DEPRECATED - The ratio of using each method to generate a structure.
                seed (int):
                    The seed for the random number generator.
                settings_out_file (str):
                    The file to write the settings to.
                verbose (int):
                    The verbosity level for the generation.
                calc (ASE calculator):
                    The calculator to use for the generated structures.
                
            Returns:
                structures (geom_rw.basis_array):
                    The generated structures.
                exit_code (int):
                    The exit code from the generation.
            """

            exit_code = 0
            structures = None

            # check if method_ratio is provided, if so, use it only if method_ratio is not provided
            if method_probab is not None and method_ratio == {"void": 0.0, "rand": 0.0, "walk": 0.0, "grow": 0.0, "min": 0.0}:
                method_ratio = method_probab
                warnings.warn("method_probab is deprecated, use method_ratio instead", DeprecationWarning)
            elif method_probab is not None:
                warnings.warn("method_probab is deprecated, use method_ratio instead", DeprecationWarning)
                # break if both method_ratio and method_probab are provided
                raise ValueError("Both method_ratio and method_probab are provided, use only one (method_ratio)")
            method_ratio_list = []
            method_ratio_list.append(method_ratio.get("void", 0.0))
            method_ratio_list.append(method_ratio.get("rand", 0.0)) # or method_ratio.get("random", 0.0))
            method_ratio_list.append(method_ratio.get("walk", 0.0))
            method_ratio_list.append(method_ratio.get("grow", 0.0)) # or method_ratio.get("growth", 0.0))
            method_ratio_list.append(method_ratio.get("min", 0.0))  # or method_ratio.get("minimum", 0.0) or method_ratio.get("global", 0.0))

            # check if all values are 0.0, if so, set them to the default of all 1.0
            if all([val < 1E-6 for val in method_ratio_list]):
                method_ratio_list = [1.0, 0.1, 0.5, 0.5, 1.0]

            # if stoichiometry is a dictionary, convert it to a stoichiometry_array
            if isinstance(stoichiometry, dict):
                stoichiometry = Generator.stoichiometry_array(dict=stoichiometry)

            exit_code = _raffle.f90wrap_generator__generate__binding__rgt(
                this=self._handle,
                num_structures=num_structures,
                stoichiometry=stoichiometry._handle,
                method_ratio=method_ratio_list,
                settings_out_file=settings_out_file,
                seed=seed, verbose=verbose)
                
            structures = self.get_structures(self, calc)[-num_structures:]
            return structures, exit_code

        def get_structures(self, calculator=None):
            """
            Get the generated structures as a list of ASE Atoms objects.

            Parameters:
                calculator (ASE calculator):
                    The calculator to use for the generated structures.
            """
            atoms = []
            for structure in self.structures:
                atoms.append(structure.toase(calculator))
            return atoms

        def set_structures(self, structures):
            """
            Set the list of generated structures.

            Parameters:
                structures (list of geom_rw.basis or list of ase.Atoms):
                    The list of structures to set.
            """
            structures = geom_rw.basis_array(atoms=structures)
            _raffle.f90wrap_generator__set_structures__binding__rgt(this=self._handle, \
                structures=structures._handle)

        def remove_structure(self, index):
            """
            Remove the structure at the given indices from the generator.

            Parameters:
                index (int or list of int):
                    The indices of the structure to remove.
            """
            index_list = [index] if isinstance(index, int) else index
            index_list = [ i + 1 for i in index_list ]
            _raffle.f90wrap_generator__remove_structure__binding__rgt(this=self._handle, \
                index_bn=index_list)

        def evaluate(self, basis):
            """
            Evaluate the viability of the structures.

            WARNING: This function is not implemented yet.

            Parameters:
                basis (geom_rw.basis or Atoms):
                    The basis to use for the evaluation.

            Returns:
                viability (float):
                    The viability of the structures.
            """
            from ase import Atoms
            if isinstance(basis, Atoms):
                basis = geom_rw.basis(atoms=basis)

            viability = \
                _raffle.f90wrap_generator__evaluate__binding__rgt(this=self._handle, \
                basis=basis._handle)
            return viability

        def print_settings(self, file):
            """
            Print the settings for the generation to a file.

            Parameters:
                file (str):
                    Name of the file to write the settings to.
            """
            _raffle.f90wrap_generator__print_settings__binding__rgt(this=self._handle, \
                file=file)

        def read_settings(self, file):
            """
            Read the settings for the generation from a file.

            Parameters:
                file (str):
                    Name of the file to read the settings from.
            """
            _raffle.f90wrap_generator__read_settings__binding__rgt(this=self._handle, \
                file=file)

        @property
        def num_structures(self):
            """
            The number of generated structures currently stored in the generator.
            """
            return _raffle.f90wrap_raffle_generator_type__get__num_structures(self._handle)

        @num_structures.setter
        def num_structures(self, num_structures):
            _raffle.f90wrap_raffle_generator_type__set__num_structures(self._handle, \
                num_structures)

        @property
        def host(self):
            """
            The host structure for the generation.
            """
            host_handle = _raffle.f90wrap_raffle_generator_type__get__host(self._handle)
            if tuple(host_handle) in self._objs:
                host = self._objs[tuple(host_handle)]
            else:
                host = geom_rw.basis.from_handle(host_handle)
                self._objs[tuple(host_handle)] = host
            return host

        @host.setter
        def host(self, host):
            host = host._handle
            _raffle.f90wrap_raffle_generator_type__set__host(self._handle, host)

        @property
        def grid(self):
            """
            The grid parameters for the generation.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_raffle_generator_type__array__grid(self._handle)
            if array_handle in self._arrays:
                grid = self._arrays[array_handle]
            else:
                grid = f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_raffle_generator_type__array__grid)
                self._arrays[array_handle] = grid
            return grid

        @grid.setter
        def grid(self, grid):
            self.grid[...] = grid

        @property
        def grid_offset(self):
            """
            The offset of the grid from the origin.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_raffle_generator_type__array__grid_offset(self._handle)
            if array_handle in self._arrays:
                grid_offset = self._arrays[array_handle]
            else:
                grid_offset = f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_raffle_generator_type__array__grid_offset)
                self._arrays[array_handle] = grid_offset
            return grid_offset

        @grid_offset.setter
        def grid_offset(self, grid_offset):
            self.grid_offset[...] = grid_offset

        @property
        def grid_spacing(self):
            """
            The spacing between grid points.
            """
            return _raffle.f90wrap_raffle_generator_type__get__grid_spacing(self._handle)

        @grid_spacing.setter
        def grid_spacing(self, grid_spacing):
            _raffle.f90wrap_raffle_generator_type__set__grid_spacing(self._handle, \
                grid_spacing)

        @property
        def bounds(self):
            """
            The bounds in direct coordinates of the host cell for the generation.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_raffle_generator_type__array__bounds(self._handle)
            if array_handle in self._arrays:
                bounds = self._arrays[array_handle]
            else:
                bounds = f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_raffle_generator_type__array__bounds)
                self._arrays[array_handle] = bounds
            return bounds

        @bounds.setter
        def bounds(self, bounds):
            self.bounds[...] = bounds

        @property
        def distributions(self):
            """
            The container for the distribution functions used in the generation.
            """
            distributions_handle = \
                _raffle.f90wrap_raffle_generator_type__get__distributions(self._handle)
            if tuple(distributions_handle) in self._objs:
                distributions = self._objs[tuple(distributions_handle)]
            else:
                distributions = \
                    raffle__distribs_container.distribs_container_type.from_handle(distributions_handle)
                self._objs[tuple(distributions_handle)] = distributions
            return distributions

        @distributions.setter
        def distributions(self, distributions):
            distributions = distributions._handle
            _raffle.f90wrap_raffle_generator_type__set__distributions(self._handle, \
                distributions)

        @property
        def max_attempts(self):
            """
            The maximum number of attempts to generate a structure using the walk method.
            """
            return _raffle.f90wrap_raffle_generator_type__get__max_attempts(self._handle)

        @max_attempts.setter
        def max_attempts(self, max_attempts):
            _raffle.f90wrap_raffle_generator_type__set__max_attempts(self._handle, \
                max_attempts)

        @property
        def walk_step_size_coarse(self):
            """
            The coarse step size for the walk method.
            """
            return \
                _raffle.f90wrap_raffle_generator_type__get__walk_step_size_coarse(self._handle)

        @walk_step_size_coarse.setter
        def walk_step_size_coarse(self, walk_step_size_coarse):
            _raffle.f90wrap_raffle_generator_type__set__walk_step_size_coarse(self._handle, \
                walk_step_size_coarse)

        @property
        def walk_step_size_fine(self):
            """
            The fine step size for the walk method.
            """
            return \
                _raffle.f90wrap_raffle_generator_type__get__walk_step_size_fine(self._handle)

        @walk_step_size_fine.setter
        def walk_step_size_fine(self, walk_step_size_fine):
            _raffle.f90wrap_raffle_generator_type__set__walk_step_size_fine(self._handle, \
                walk_step_size_fine)

        @property
        def method_ratio(self):
            """
            The ratio of methods to employ to generate a structure.
            """
            array_ndim, array_type, array_shape, array_handle = \
                _raffle.f90wrap_raffle_generator_type__array__method_ratio(self._handle)
            if array_handle in self._arrays:
                method_ratio = self._arrays[array_handle]
            else:
                method_ratio = f90wrap.runtime.get_array(f90wrap.runtime.sizeof_fortran_t,
                                        self._handle,
                                        _raffle.f90wrap_raffle_generator_type__array__method_ratio)
                self._arrays[array_handle] = method_ratio
            return method_ratio

        @method_ratio.setter
        def method_ratio(self, method_ratio):
            self.method_ratio[...] = method_ratio

        def _init_array_structures(self):
            """
            Initialise the structures array.

            It is not recommended to use this function directly. Use the `structures` property instead.
            """
            self.structures = f90wrap.runtime.FortranDerivedTypeArray(self,
                                            _raffle.f90wrap_raffle_generator_type__array_getitem__structures,
                                            _raffle.f90wrap_raffle_generator_type__array_setitem__structures,
                                            _raffle.f90wrap_raffle_generator_type__array_len__structures,
                                            """
            Element items ftype=type(basis_type) pytype=basis


            Defined at ../src/lib/mod_generator.f90 line \
                29

            """, Geom_Rw.basis)
            return self.structures

        def __str__(self):
            ret = ['<raffle_generator>{\n']
            ret.append('    num_structures : ')
            ret.append(repr(self.num_structures))
            ret.append(',\n    host : ')
            ret.append(repr(self.host))
            ret.append(',\n    grid : ')
            ret.append(repr(self.grid))
            ret.append(',\n    grid_offset : ')
            ret.append(repr(self.grid_offset))
            ret.append(',\n    grid_spacing : ')
            ret.append(repr(self.grid_spacing))
            ret.append(',\n    bounds : ')
            ret.append(repr(self.bounds))
            ret.append(',\n    distributions : ')
            ret.append(repr(self.distributions))
            ret.append(',\n    max_attempts : ')
            ret.append(repr(self.max_attempts))
            ret.append(',\n    method_ratio : ')
            ret.append(repr(self.method_ratio))
            ret.append(',\n    structures : ')
            ret.append(repr(self.structures))
            ret.append('}')
            return ''.join(ret)

        _dt_array_initialisers = [_init_array_structures]


    _dt_array_initialisers = []


generator = Generator()

class Raffle(f90wrap.runtime.FortranModule):
    """
    Module raffle


    Defined at ../src/raffle.f90 lines 1-4

    """
    pass
    _dt_array_initialisers = []


raffle = Raffle()



import os
import glob
from raffle.generator import raffle_generator
import pytest
from ase.io import read
import time
import numpy as np


def get_xyz_files():
    """Get all .xyz files from the ../../data directory"""
    current_dir = os.path.dirname(os.path.abspath(__file__))
    data_dir = os.path.join(current_dir, '..', '..', '..', 'example', 'data')
    print("Data directory:", data_dir)
    return glob.glob(os.path.join(data_dir, '*.xyz'))


@pytest.mark.parametrize("xyz_file", get_xyz_files())
@pytest.mark.parametrize("kBT", [0.2, 0.4, 0.6])
@pytest.mark.parametrize("width", [
    [0.025, np.pi/200.0, np.pi/200.0],
    [0.05, np.pi/100.0, np.pi/100.0],
    [0.1, np.pi/50.0, np.pi/50.0]
])
@pytest.mark.benchmark(
    group="distributions",
    min_rounds=5,
    timer=time.time,
    disable_gc=True,
    warmup=False
)
def test_distributions_create(benchmark, xyz_file, kBT, width):
    atoms = read(xyz_file, index=":")
    
    # Initialize generator
    generator = raffle_generator()

    # get list of elements in list of atoms
    elements = []
    for atom in atoms:
        elements += atom.get_chemical_symbols()
    elements = list(set(elements))
    energies = {element: 0.0 for element in elements}
    generator.distributions.set_element_energies(energies)

    avg_num_atoms = 0
    for atom in atoms:
        avg_num_atoms += len(atom)
    avg_num_atoms = round(avg_num_atoms / len(atoms))

    benchmark.extra_info = {
        'file': os.path.basename(xyz_file),
        'num_structures': len(atoms),
        'num_species': len(elements),
        "avg_num_atoms": avg_num_atoms,
        "kBT": kBT,
        "width": width
    }

    # set energy scale
    generator.distributions.set_kBT(kBT)

    # set the distribution function widths (2-body, 3-body, 4-body)
    generator.distributions.set_width(width)
    
    # Measure time for distributions.create()
    result = benchmark(generator.distributions.create, atoms)

    
    return {
        'file': os.path.basename(xyz_file),
        'time': result,
        'num_structures': len(atoms),
        'num_species': len(elements),
        "avg_num_atoms": avg_num_atoms
    }

@pytest.fixture
def make_customer_record():
    def _make_customer_record(name):
        return {"name": name, "orders": []}

    return _make_customer_record

if __name__ == '__main__':
    xyz_files = get_xyz_files()
    results = []
    
    result = test_distributions_create(lambda x: x)
    # for xyz_file in xyz_files:
    #     # for _ in range(5):  # Run the benchmark 5 times for each file
    #     result = test_distributions_create(lambda x: x, xyz_file)
    #     results.append(result)
    
    # Print results
    print("\nBenchmark Results:")
    print("-" * 60)
    print(f"{'File':<20} {'Time (s)':<10} {'Num Structures':<10} {'Num Species':<10} {'Avg Num Atoms':<10}")
    print("-" * 60)
    for result in results:
        print(f"{result['file']:<20} {result['time']:<10.4f} {result['num_structures']:<10} {result['num_species']:<10} {result['avg_num_atoms']:<10}")


import pytest
import time
import numpy as np
from ase import Atoms
from chgnet.model import CHGNetCalculator
from raffle.generator import raffle_generator


@pytest.mark.parametrize("kBT", [0.4])
@pytest.mark.parametrize("width", [
    [0.025, np.pi/200.0, np.pi/200.0]
])
@pytest.mark.parametrize("grid_spacing", [0.5, 0.25, 0.1, 0.075, 0.05])
@pytest.mark.parametrize("method_ratio", [
  {'void': 0.0, 'rand': 0.0, 'walk': 0.0, 'grow': 0.0, 'min': 1.0},
  {'void': 0.0, 'rand': 0.0, 'walk': 0.0, 'grow': 1.0, 'min': 0.0},
  {'void': 0.0, 'rand': 0.0, 'walk': 1.0, 'grow': 0.0, 'min': 0.0},
  {'void': 0.0, 'rand': 1.0, 'walk': 0.0, 'grow': 0.0, 'min': 0.0},
  {'void': 1.0, 'rand': 0.0, 'walk': 0.0, 'grow': 0.0, 'min': 0.0}
])
@pytest.mark.benchmark(
    group="placement_methods",
    min_rounds=5,
    timer=time.time,
    disable_gc=True,
    warmup=False
)
def test_placement_methods(benchmark, grid_spacing, kBT, width, method_ratio):
    
    calc = CHGNetCalculator()
    host = Atoms('C', positions=[(0, 0, 0)], cell=[3.567, 3.567, 3.567], pbc=True, calculator=calc)

    # Initialize generator
    generator = raffle_generator()

    # set energy scale
    generator.distributions.set_kBT(kBT)

    # set the distribution function widths (2-body, 3-body, 4-body)
    generator.distributions.set_width(width)

    energies = {element: 0.0 for element in ['C']}
    generator.distributions.set_element_energies(energies)

    generator.set_grid(grid_spacing=grid_spacing)
    generator.set_host(host)

    initial_database = [Atoms('C', positions=[(0, 0, 0)], cell=[8, 8, 8], pbc=True, calculator=calc)]

    generator.distributions.create(initial_database)
    
    benchmark.extra_info = {
        "kBT": kBT,
        "width": width,
        "grid_spacing": grid_spacing,
        "method_ratio": method_ratio
    }

    # Measure time for distributions.create()
    result = benchmark(generator.generate, 
        num_structures = 1,
        stoichiometry = {'C': 1},
        seed = 0,
        method_ratio = method_ratio,
        verbose = 0
    )

    return result

if __name__ == '__main__':
    grid_spacings = [0.5, 0.25, 0.1, 0.075, 0.05]
    results = []

    for grid_spacing in grid_spacings:
        for _ in range(5):  # Run the benchmark 5 times for each grid size
            result = test_placement_methods(lambda x: x, grid_spacing, 0.4, [0.025, np.pi/200.0, np.pi/200.0])
            results.append({
                'grid_spacing': grid_spacing,
                'time': result
            })
    
    # Print results
    print("\nBenchmark Results:")
    print("-" * 40)
    print(f"{'Grid Size':<10} {'Time (s)':<10}")
    print("-" * 40)
    for result in results:
        print(f"{result['grid_spacing']:<10} {result['time']:<10.4f}")



import unittest
from parameterized import parameterized
import os
from raffle.raffle import Geom_Rw, Raffle__Distribs_Container, Generator
import raffle
from ase import Atoms
import subprocess



fc = os.getenv("FORTRAN_COMPILER")
temp_test_dir = os.path.join("build", ".temp_test")
dirname = os.path.dirname(raffle.__file__)
include_dir = os.path.join(dirname, "include")
etc_dir = os.path.join(dirname, "etc")
lib_dir = os.path.join(dirname, "lib")

with open("test/CMakeLists.txt", "r") as file:
    lines = file.readlines()

test_names = []
read_names = False
for line in lines:
    if line.startswith("foreach(execid"):
        read_names = True
    elif line.strip().startswith(")"):
        read_names = False
        break
    elif read_names:
        print(line)
        test_name = line.split()[0]
        test_names.append(test_name)

def build_fortran_test(test_name):
    os.makedirs(temp_test_dir, exist_ok=True)
    # make using a fortran compiler
    # point to raffle's pip install include and lib to include during compilation

    build_result = subprocess.run(
        [
            fc, "-o", test_name+".o",
            "../../test/test_io_utils.f90",
            "-I", include_dir,
            "-I", etc_dir,
            "-L", lib_dir,
            "-lraffle"
        ],
        cwd=temp_test_dir
    )
    return build_result

def run_fortran_test(test_name):
    run_result = subprocess.run(
        ["./"+test_name+".o"],
        cwd=temp_test_dir
    )
    return run_result


class TestGeomRw(unittest.TestCase):

    def test_species_type_initialization(self):
        species = Geom_Rw.species_type()
        self.assertIsNotNone(species._handle)

    def test_species_type_properties(self):
        species = Geom_Rw.species_type()
        species.mass = 12.0
        self.assertEqual(species.mass, 12.0)
        species.charge = 1.0
        self.assertEqual(species.charge, 1.0)
        species.radius = 0.5
        self.assertEqual(species.radius, 0.5)
        species.name = "C"
        self.assertEqual(species.name, b"C")
        species.num = 2
        self.assertEqual(species.num, 2)

    def test_basis_initialization(self):
        basis = Geom_Rw.basis()
        self.assertIsNotNone(basis._handle)

    def test_basis_properties(self):
        basis = Geom_Rw.basis()
        basis.nspec = 2
        self.assertEqual(basis.nspec, 2)
        basis.natom = 4
        self.assertEqual(basis.natom, 4)
        basis.energy = -10.0
        self.assertEqual(basis.energy, -10.0)
        basis.lat = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
        self.assertTrue((basis.lat == [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]).all())
        basis.lcart = True
        self.assertTrue(basis.lcart)
        basis.pbc = [True, True, True]
        self.assertTrue((basis.pbc == [True, True, True]).all())
        basis.sysname = "Test"
        self.assertEqual(basis.sysname, b"Test")

    def test_basis_array_initialization(self):
        basis_array = Geom_Rw.basis_array()
        self.assertIsNotNone(basis_array._handle)

    def test_basis_array_allocate(self):
        basis_array = Geom_Rw.basis_array()
        basis_array.allocate(2)
        self.assertEqual(len(basis_array.items), 2)

class TestRaffleDistribsContainer(unittest.TestCase):

    def test_distribs_container_initialization(self):
        distribs_container = Raffle__Distribs_Container.distribs_container_type()
        self.assertIsNotNone(distribs_container._handle)

    def test_distribs_container_set_kBT(self):
        distribs_container = Raffle__Distribs_Container.distribs_container_type()
        distribs_container.set_kBT(0.025)
        self.assertTrue( abs(distribs_container.kBT - 0.025) < 1e-6 )

    def test_distribs_container_set_weight_method(self):
        distribs_container = Raffle__Distribs_Container.distribs_container_type()
        distribs_container.set_weight_method('formation_energy')
        self.assertFalse(distribs_container.weight_by_hull)
        distribs_container.set_weight_method('energy_above_hull')
        self.assertTrue(distribs_container.weight_by_hull)

    def test_distribs_container_set_width(self):
        distribs_container = Raffle__Distribs_Container.distribs_container_type()
        distribs_container.set_width([0.1, 0.2, 0.3])
        self.assertTrue( ( abs( distribs_container.width - [0.1, 0.2, 0.3] ) < 1e-6 ).all() )

    def test_distribs_container_set_sigma(self):
        distribs_container = Raffle__Distribs_Container.distribs_container_type()
        distribs_container.set_sigma([0.1, 0.2, 0.3])
        self.assertTrue( ( abs( distribs_container.sigma - [0.1, 0.2, 0.3] ) < 1e-6 ).all() )

    def test_distribs_container_set_cutoff_min(self):
        distribs_container = Raffle__Distribs_Container.distribs_container_type()
        distribs_container.set_cutoff_min([0.1, 0.2, 0.3])
        self.assertTrue( ( abs( distribs_container.cutoff_min - [0.1, 0.2, 0.3] ) < 1e-6 ).all() )

    def test_distribs_container_set_cutoff_max(self):
        distribs_container = Raffle__Distribs_Container.distribs_container_type()
        distribs_container.set_cutoff_max([0.1, 0.2, 0.3])
        self.assertTrue( ( abs( distribs_container.cutoff_max - [0.1, 0.2, 0.3] ) < 1e-6 ).all() )

    def test_distribs_container_set_radius_distance_tol(self):
        distribs_container = Raffle__Distribs_Container.distribs_container_type()
        distribs_container.set_radius_distance_tol([0.1, 0.2, 0.3, 0.4])
        self.assertTrue( ( abs(distribs_container.radius_distance_tol - [0.1, 0.2, 0.3, 0.4]) < 1e-6 ).all() )

class TestGenerator(unittest.TestCase):

    def test_generator_initialization(self):
        generator = Generator.raffle_generator()
        self.assertIsNotNone(generator._handle)

    def test_set_max_attempts(self):
        generator = Generator.raffle_generator()
        generator.set_max_attempts(10)
        self.assertEqual(generator.max_attempts, 10)

    def test_set_walk_step_size(self):
        generator = Generator.raffle_generator()
        generator.set_walk_step_size(coarse=0.5, fine=0.1)
        self.assertTrue( abs( generator.walk_step_size_coarse - 0.5 ) < 1e-6 )
        self.assertTrue( abs( generator.walk_step_size_fine - 0.1 ) < 1e-6 )

    def test_set_host(self):
        generator = Generator.raffle_generator()
        atoms = Atoms('H2', positions=[[0, 0, 0], [0, 0, 0.74]])
        generator.set_host(atoms)
        self.assertIsNotNone(generator.host)

    def test_set_grid(self):
        generator = Generator.raffle_generator()
        generator.set_grid(grid=[10, 10, 10], grid_offset=[0.1, 0.2, 0.3])
        self.assertTrue( ( generator.grid == [10, 10, 10] ).all() )
        self.assertTrue( ( abs( generator.grid_offset - [0.1, 0.2, 0.3] ) < 1e-6 ).all() )

        generator.set_grid(grid_spacing=0.1)
        self.assertTrue( abs( generator.grid_spacing - 0.1 ) < 1e-6 )

    def test_reset_grid(self):
        generator = Generator.raffle_generator()
        generator.set_grid(grid=[10, 10, 10], grid_offset=[0.0, 0.0, 0.0])
        generator.reset_grid()
        self.assertTrue((generator.grid == [0, 0, 0]).all())

    def test_set_bounds(self):
        generator = Generator.raffle_generator()
        bounds = [ [ 0.1, 0.2, 0.3 ], [0.4, 0.5, 0.6 ] ]
        generator.set_bounds(bounds)
        self.assertTrue( ( abs( generator.bounds - bounds ) < 1e-6 ).all() )

    def test_reset_bounds(self):
        generator = Generator.raffle_generator()
        bounds = [ [ 0.1, 0.2, 0.3 ], [0.4, 0.5, 0.6 ] ]
        generator.set_bounds(bounds)
        generator.reset_bounds()
        self.assertTrue( ( abs( generator.bounds - [ [ 0.0, 0.0, 0.0 ], [ 1.0, 1.0, 1.0 ] ] ) < 1e-6 ).all() )

    # def test_generate(self):
    #     generator = Generator.raffle_generator()
    #     atoms = Atoms('H2', positions=[[0, 0, 0], [0, 0, 0.74]], cell=[10, 10, 10])
    #     generator.set_host(atoms)
    #     generator.distributions.set_element_energies( {"H": 0.0} )
    #     stoichiometry = {"H": 2}
    #     generator.generate(num_structures=1, stoichiometry=stoichiometry)
    #     self.assertEqual(generator.num_structures, 1)

    def test_get_structures(self):
        generator = Generator.raffle_generator()
        atoms = Atoms('H2', positions=[[0, 0, 0], [0, 0, 0.74]], cell=[10, 10, 10])
        generator.set_structures([atoms])
        structures = generator.get_structures()
        self.assertEqual( len(structures), 1 )

    def test_set_structures(self):
        generator = Generator.raffle_generator()
        atoms = Atoms('H2', positions=[[0, 0, 0], [0, 0, 0.74]], cell=[10, 10, 10])
        generator.set_structures([atoms])
        self.assertEqual(generator.num_structures, 1)

    def test_remove_structure(self):
        generator = Generator.raffle_generator()
        atoms = []
        atoms.append(Atoms('H2', positions=[[0, 0, 0], [0, 0, 0.74]], cell=[10, 10, 10]))
        atoms.append(Atoms('H2', positions=[[0, 0, 0], [0, 0, 0.74]], cell=[10, 10, 10]))
        generator.set_structures(atoms)
        generator.remove_structure(0)
        self.assertEqual(generator.num_structures, 1)

    # def test_evaluate(self):
    #     generator = Generator.raffle_generator()
    #     atoms = Atoms('H2', positions=[[0, 0, 0], [0, 0, 0.74]], cell=[10, 10, 10])
    #     generator.set_host(atoms)
    #     generator.distributions.set_element_energies( {"H": 0.0} )
    #     generator.distributions.create([atoms])
    #     basis = Geom_Rw.basis(atoms)
    #     viability = generator.evaluate(basis)
    #     self.assertIsNotNone(viability)

    def test_print_settings(self):
        generator = Generator.raffle_generator()
        atoms = Atoms('H2', positions=[[0, 0, 0], [0, 0, 0.74]], cell=[10, 10, 10])
        generator.set_host(atoms)
        generator.distributions.set_element_energies( {"H": 0.0} )
        generator.distributions.create([atoms])

        # check if filename exists
        i = 0
        filename = ".raffle_generator_settings.txt"
        while os.path.isfile(filename):
            i += 1
            filename = ".raffle_generator_settings"+str(i)+".txt"

        generator.print_settings(filename)
        with open(filename, "r") as file:
            settings = file.read()
        self.assertTrue(len(settings) > 0)
        os.remove(filename)

class TestFortranUnits(unittest.TestCase):

    @parameterized.expand([(test_name) for test_name in test_names])
    def test_fortran(self, test_name):
        build_result = build_fortran_test(test_name)
        self.assertEqual(build_result.returncode, 0)
        run_result = run_fortran_test(test_name)
        self.assertEqual(run_result.returncode, 0)

if __name__ == '__main__':
    unittest.main()


"""
raffle package

This package provides functionality to interface with a Fortran library,
including a Python wrapper around the Fortran code.
"""

from importlib.metadata import PackageNotFoundError, version
try:
    __version__ = version(__name__)
except PackageNotFoundError:
    __version__ = "unknown"

from .raffle import generator as _generator_class
from .raffle import geom_rw as _geom_rw_class
# from .raffle import generator


# Use the 'types' module to create simulated 'generator' and 'geom submodules
import types
generator = types.ModuleType('generator')
geom = types.ModuleType('geom')

# Assign the respective class to the simulated 'generator' and 'geom' modules
generator.raffle_generator = _generator_class.raffle_generator
generator.stoichiometry_array = _generator_class.stoichiometry_array

# Assign the class to the simulated 'geom' module
geom.basis_array = _geom_rw_class.basis_array
geom.basis = _geom_rw_class.basis


# Add the simulated 'generator' and 'geom' module to the current package
import sys
sys.modules['raffle.generator'] = generator
sys.modules['raffle.geom'] = geom

# Clean up internal imports (remove access to the direct classes)
del _generator_class
del _geom_rw_class
del PackageNotFoundError
del version
del sys
del types
del raffle

__all__ = ['__version__', 'generator', 'geom']

def __getattr__(name):
    if name == "generator":
        return generator
    elif name == "geom":
        return geom
    raise AttributeError(f"module {__name__} has no attribute {name}")


# %%
#Imports MPRester
# from mp_api.client import MPRester
from ase import Atoms
from ase.io import write, read
from pymatgen.io.ase import AseAtomsAdaptor
from mp_api.client import MPRester


# %%
# Personal api key for accessing materials project api
# This is unique to each user
# api_key = ""

# %%
#Gets the ID of materials we want. Ba is all structures with only Ba. Ba-O is all structures with only Ba and O, not only BaO.

mpr = MPRester() # MPRester(api_key)
materials = []

materials.append(mpr.materials.summary.search(chemsys="C", 
                            fields=["material_id","structure", "energy_per_atom", "nsites"]))
materials.append(mpr.materials.summary.search(chemsys="Mg", 
                            fields=["material_id","structure", "energy_per_atom", "nsites"]))
materials.append(mpr.materials.summary.search(chemsys="O", 
                            fields=["material_id","structure", "energy_per_atom", "nsites"]))
materials.append(mpr.materials.summary.search(chemsys="C-Mg", 
                            fields=["material_id","structure", "energy_per_atom", "nsites"]))
materials.append(mpr.materials.summary.search(chemsys="C-O", 
                            fields=["material_id","structure", "energy_per_atom", "nsites"]))
materials.append(mpr.materials.summary.search(chemsys="Mg-O", 
                            fields=["material_id","structure", "energy_per_atom", "nsites"]))


# %%
structures = []
energies = []
nsites = []
for material_set in materials:
    for material in material_set:
        material_id = material.material_id
        structures.append(mpr.get_structure_by_material_id(material_id))
        energies.append(material.energy_per_atom)
        nsites.append(material.nsites)

# %%
all_atoms = []
for structure, energy, nsite in zip(structures, energies, nsites):
    atom = AseAtomsAdaptor.get_atoms(structure)
    atom.info['free_energy'] = energy * nsite
    atom.info['energy'] = energy * nsite
    all_atoms.append(atom)
write("database.xyz", all_atoms, format='extxyz')

# %%






# This script demonstrates how to use the raffle generator to generate perovskite supperlattice structures

# The script reads a host structure from a POSCAR file, and sets it as the host structure for the generator.

# import standard libraries
import os
import numpy as np

# import raffle library
from raffle.generator import raffle_generator

# import ASE (Atomic Simulation Environment) modules
from ase import Atoms
from ase.io import read, write
from ase.optimize import BFGS

# import CHGNet calculator (for MLP energy evaluation)
from chgnet.model.dynamics import CHGNetCalculator

# load CHGNet calculator
print("Initialising CHGNet calculator")
calculator = CHGNetCalculator(model=None)

# set up an instance of the raffle generator
print("Initialising raffle generator")
generator = raffle_generator()

# read the host structure from a POSCAR file
print("Reading host")
host = read("POSCAR_host")
generator.set_host(host)
print("Host read")

host.calc = calculator
print("host energy: ", host.get_potential_energy())

# set the element reference energies
print("Setting element energies")
generator.distributions.set_element_energies(
    {
        'Ba': -1.9030744,
        'Sr': -1.5478236,
        'Ti': -7.842818,
        'O': -4.3707458
    }
)

# set the distribution function widths (2-body, 3-body, 4-body)
print("Reading database")
database = read("database.xyz", index=":")


# create the distribution functions
print("Setting database")
generator.distributions.create(database, deallocate_systems=False)
print("Database set")

# print the distribution functions to a file
print("Printing distributions")
generator.distributions.write("distributions.txt")
generator.distributions.write_2body("df2.txt")
generator.distributions.write_3body("df3.txt")
generator.distributions.write_4body("df4.txt")

# check the element energies and bond radii
print("Checking element energies")
print(generator.distributions.get_element_energies())
print("Checking bond radii")
print(generator.distributions.get_bond_radii())

# set the grid for the host cell
print("Setting bins (discretisation of host cell)")
generator.set_grid(grid=[8,8,128], grid_offset=[0.0, 0.0, 0.0])
print(generator.grid)

# set the stoichiometry for the structures to be generated
print("Setting stoichiometry to insert")
stoich_dict = { 'Ba': 4, 'Sr': 4, 'Ti': 8, 'O': 24 }

# generate structures
num_structures_old = 0
optimise_structure = True
for iter in range(20):
    print(f"Iteration {iter}")
    print("Generating...")
    # this is the main function to generate structures
    generator.generate(num_structures=1, stoichiometry=stoich_dict, seed=0+iter, verbose=0, method_ratio={"void":0.001, "walk":0.0, "min":1.0})
    print("Generated")

    print("Getting structures")
    print("number of structures supposed to be generated: ", generator.num_structures)
    generated_structures = generator.get_structures()
    print("actual number allocated: ",len(generated_structures))
    print("Got structures")

    # check if directory iteration[iter] exists, if not create it
    iterdir = f"iteration{iter}/"
    if not os.path.exists(iterdir):
        os.makedirs(iterdir)

    # get energies using MLPs and optimise the structures
    num_structures_new = len(generated_structures)
    for i, atoms in enumerate(generated_structures):
        if(i < num_structures_old):
            continue
        inew = i - num_structures_old
        atoms.calc = calculator
        if optimise_structure:
            optimizer = BFGS(atoms, trajectory = "traje.traj")
            optimizer.run(fmax=0.05)
            print(f"Structure {inew} optimised")
        atoms.get_potential_energy()
        print(f"Structure {inew} energy: {atoms.get_potential_energy()}")
        write(iterdir+f"POSCAR_{inew}", atoms)

    # update the distribution functions
    print("Updating distributions")
    generator.distributions.update(generated_structures[num_structures_old:], deallocate_systems=False)

    # print the new distribution functions to a file
    print("Printing distributions")
    generator.distributions.write(iterdir+"distributions.txt")
    generator.distributions.write_2body(iterdir+"df2.txt")
    generator.distributions.write_3body(iterdir+"df3.txt")
    generator.distributions.write_4body(iterdir+"df4.txt")

    # update the number of structures generated
    num_structures_old = num_structures_new



# from mace.calculators import mace_mp
# from ase.calculators.vasp import Vasp
from chgnet.model import CHGNetCalculator
from raffle.generator import raffle_generator
from ase import build, Atoms
from ase.optimize import BFGS, FIRE
from ase.io import write
from ase.visualize import view
import numpy as np
import os
from concurrent.futures import ProcessPoolExecutor, wait, as_completed
from multiprocessing import Process
from copy import deepcopy
from multiprocessing import Queue
from joblib import Parallel, delayed

import logging
logging.basicConfig(level=logging.DEBUG)

def runInParallel(*fns):
    proc = []
    results = []
    for fn in fns:
        p = Process(target=fn)
        p.start()
        proc.append(p)
    for p in proc:
        results.append(p.join())

    print("All processes finished")
    print(results)


def process_structure_with_queue(i, structure, num_old, calc_params, optimise_structure, iteration, queue):
    # Perform the computation
    result = process_structure(i, structure, num_old, calc_params, optimise_structure, iteration)
    queue.put(result)  # Report completion

def process_structure(i, atoms, num_structures_old, calc_params, optimise_structure, iteration):
    if i < num_structures_old:
        return
    
    # calc = Vasp(**calc_params, label=f"struct{i}", directory=f"iteration{iteration}/struct{i}/", txt=f"stdout{i}.o")
    inew = i - num_structures_old
    atoms.calc = calc
    # positions_initial = atoms.get_positions()

    # Calculate and save the initial energy per atom
    energy_unrlxd = atoms.get_potential_energy() / len(atoms)
    print(f"Initial energy per atom: {energy_unrlxd}")

    # Optimise the structure if requested
    if optimise_structure:
        optimizer = FIRE(atoms, trajectory = f"traje{inew}.traj", logfile=f"optimisation{inew}.log")
        try:
            optimizer.run(fmax=0.05, steps=100)
        except Exception as e:
            print(f"Optimisation failed: {e}")
            return None, None, None
    
    # Save the optimised structure and its energy per atom
    energy_rlxd = atoms.get_potential_energy() / len(atoms)

    # Get the distance matrix
    distances = atoms.get_all_distances(mic=True)

    # Set self-distances (diagonal entries) to infinity to ignore them
    np.fill_diagonal(distances, np.inf)

    # Check if the minimum non-self distance is below 1.5
    if distances.min() < 1.0:
        print(f"Distance too small: {atoms.get_all_distances(mic=True).min()}")
        return None, None, None
    
    if abs(energy_rlxd - energy_unrlxd) > 10.0:
        print(f"Energy difference too large: {energy_rlxd} vs {energy_unrlxd}")
        return None, None, None
    
    return atoms, energy_unrlxd, energy_rlxd

# crystal_structures = [
#     'sc', 'fcc', 'bcc', 'hcp',
#     'diamond', 'zincblende', 'rocksalt', 'cesiumchloride',
#     'fluorite', 'wurtzite', 'tetragonal', 'orthorhombic',
#     'bct', 'rhombohedral', 'mcl'
# ]



if __name__ == "__main__":

    calc_params = {}
    calc = CHGNetCalculator()
    # calc_params = {
    #     "model": "large",
    #     "dispersion": False,
    #     "default_dtype": "float32",
    #     "device": 'cpu'
    # }
    # calc = mace_mp(**calc_params)
    # calc_params = {
    #     "command": "$HOME/DVASP/vasp.6.4.3/bin/vasp_std",
    #     # "label": "iteration",
    #     # "txt": "stdout.o",
    #     "xc": 'pbe',
    #     "setups": {'C': ''},
    #     "kpts": (3, 3, 3),
    #     "encut": 400,
    #     "istart": 0,
    #     "icharg": 0,
    # }
    # ## DON'T FORGET TO EXPORT THE VASP_PP_PATH
    # calc = Vasp(**calc_params, label="tmp", directory="tmp", txt="stdout.o")

    #crystal_structures = [
    #    'orthorhombic', 'diamond',
    #    'bct', 'sc',
    #    'fcc', 'bcc', 'hcp',
    #]

    """
    This needs to be changed to set the material cells that will be searched through.
    The crystal_structures list contains the crystal structures that will be used to generate the hosts.
    """
    crystal_structures = [
        'orthorhombic', 'hcp',
    ]

    """
    This needs to be changed to set the lattice constants that will be searched through.
    The lattice_constants list contains the lattice constants that will be used to generate the hosts.
    """
    lattice_constants = np.linspace(3.1, 5.4, num=6)

    """
    This needs to be changed to set the values of the method_ratio.
    """
    #void_val; rand_val; walk_val; grow_val; min_val
    ##############[  V,   R,   W,   G,   M]    
    method_val = [85.0, 15.0, 0.0, 0.0, 0.0] #change



    hosts = []
    for crystal_structure in crystal_structures:
        print(f'Crystal structure: {crystal_structure}')
        for a in lattice_constants:
            b = a
            c = a
            atom = build.bulk(
                    name = 'Al',
                    crystalstructure = crystal_structure,
                    a = a,
                    b = b,
                    c = c,
            )
            hosts.append(Atoms('Al', positions=[(0, 0, 0)], cell=atom.get_cell(), pbc=True, calculator=calc))
            # hosts[-1].set_pbc(True)
            # hosts[-1].calc = calc
            print(hosts[-1])

    print("number of hosts: ", len(hosts))

    optimise_structure = True
    mass = 26.9815385
    density = 1.61 # u/A^3
    # num_atoms = 7

    for seed in range(1):
        print(f"Seed: {seed}")
        energies_rlxd_filename = f"energies_rlxd_seed{seed}.txt"
        energies_unrlxd_filename = f"energies_unrlxd_seed{seed}.txt"
        generator = raffle_generator()
        generator.distributions.set_element_energies(
            {
                'Al': 0.0
            }
        )
        # set energy scale
        generator.distributions.set_kBT(0.4)
        # set the distribution function widths (2-body, 3-body, 4-body)
        generator.distributions.set_width([0.025, np.pi/200.0, np.pi/200.0])

        initial_database = [Atoms('Al', positions=[(0, 0, 0)], cell=[8, 8, 8], pbc=True)]
        initial_database[0].calc = calc

        generator.distributions.create(initial_database)

        if os.path.exists(energies_rlxd_filename):
            with open(energies_rlxd_filename, "w") as energy_file:
                pass
        else:
            open(energies_rlxd_filename, "w").close()

        if os.path.exists(energies_unrlxd_filename):
            with open(energies_unrlxd_filename, "w") as energy_file:
                pass
        else:
            open(energies_unrlxd_filename, "w").close()


        num_structures_old = 0
        unrlxd_structures = []
        rlxd_structures = []
        iter2 = 0
        for iter in range(10):
            for host in hosts:
                generator.set_host(host)
                volume = host.get_volume()

                num_atoms = round(density * volume / mass) - 1
                if(num_atoms < 1):
                    continue
                iter2 += 1
                print(f"Volume: {volume}")
                print(f"Number of atoms: {num_atoms}")
    
                generator.generate(
                    num_structures = 5,
                    stoichiometry = { 'Al': num_atoms },
                    seed = seed*1000+iter,
                    method_ratio = {"void": method_val[0], "rand": method_val[1], "walk": method_val[2], "grow": method_val[3], "min": method_val[4]},
                    verbose = 0,
                )

                # print the number of structures generated
                print("Total number of structures generated: ", generator.num_structures)
                generated_structures = generator.get_structures(calc)
                num_structures_new = len(generated_structures)

                # check if directory iteration[iter] exists, if not create it
                iterdir = f"iteration{iter2}/"
                if not os.path.exists(iterdir):
                    os.makedirs(iterdir)
                generator.print_settings(iterdir+"generator_settings.txt")

                # set up list of energies
                energy_unrlxd = np.zeros(num_structures_new - num_structures_old)
                energy_rlxd = np.zeros(num_structures_new - num_structures_old)
                for i in range(num_structures_new - num_structures_old):
                    write(iterdir+f"POSCAR_unrlxd_{i}", generated_structures[num_structures_old + i])
                    print(f"Structure {i} energy per atom: {generated_structures[num_structures_old + i].get_potential_energy() / len(generated_structures[num_structures_old + i])}")
                    unrlxd_structures.append(deepcopy(generated_structures[num_structures_old + i]))
                
                # Start parallel execution
                print("Starting parallel execution")
                results = Parallel(n_jobs=5)(
                    delayed(process_structure)(i, deepcopy(generated_structures[i]), num_structures_old, calc_params, optimise_structure, iteration=seed)
                    for i in range(num_structures_old, num_structures_new)
                )

                # Wait for all futures to complete
                for j, result in enumerate(results):
                    generated_structures[j+num_structures_old], energy_unrlxd[j], energy_rlxd[j] = result
                    rlxd_structures.append(deepcopy(generated_structures[j+num_structures_old]))
                print("All futures completed")

                # Remove structures that failed the checks
                for j, atoms in reversed(list(enumerate(generated_structures))):
                    if atoms is None:
                        energy_unrlxd = np.delete(energy_unrlxd, j-num_structures_old)
                        energy_rlxd = np.delete(energy_rlxd, j-num_structures_old)
                        del generated_structures[j]
                        # del unrlxd_structures[j]
                        del rlxd_structures[j]
                        generator.remove_structure(j)
                num_structures_new = len(generated_structures) 

                # write the structures to files
                for i in range(num_structures_new - num_structures_old):
                    write(iterdir+f"POSCAR_{i}", generated_structures[num_structures_old + i])
                    print(f"Structure {i} energy per atom: {energy_rlxd[i]}")
                    # append energy per atom to the 'energies_unrlxd_filename' file
                    with open(energies_unrlxd_filename, "a") as energy_file:
                        energy_file.write(f"{i+num_structures_old} {energy_unrlxd[i]}\n")
                    # append energy per atom to the 'energies_rlxd_filename' file
                    with open(energies_rlxd_filename, "a") as energy_file:
                        energy_file.write(f"{i+num_structures_old} {energy_rlxd[i]}\n")

                # update the distribution functions
                print("Updating distributions")
                generator.distributions.update(generated_structures[num_structures_old:], from_host=False, deallocate_systems=False)

                # print the new distribution functions to a file
                print("Printing distributions")
                generator.distributions.write_dfs(iterdir+"distributions.txt")
                generator.distributions.write_2body(iterdir+"df2.txt")
                generator.distributions.write_3body(iterdir+"df3.txt")
                generator.distributions.write_4body(iterdir+"df4.txt")
                generator.distributions.deallocate_systems()

                # update the number of structures generated
                num_structures_old = num_structures_new

        generator.distributions.write_gdfs(f"gdfs_seed{seed}.txt")

        # Read energies from the file
        with open(energies_rlxd_filename, "r") as energy_file:
            energies = energy_file.readlines()

        # Parse and sort the energies
        energies = [line.strip().split() for line in energies]
        energies = sorted(energies, key=lambda x: float(x[1]))

        # Write the sorted energies back to the file
        with open(f"sorted_{energies_rlxd_filename}", "w") as energy_file:
            for entry in energies:
                energy_file.write(f"{int(entry[0])} {float(entry[1])}\n")

        write(f"unrlxd_structures_seed{seed}.traj", unrlxd_structures)
        write(f"rlxd_structures_seed{seed}.traj", rlxd_structures)
        print("All generated and relaxed structures written")

    print("Learning complete")


import os
import re
import sys
import glob


def check_if_match(actual_indent, expected_indent, continued_indent, continuation_line, line_num, file_path):
    if continuation_line:
        if actual_indent != continued_indent:
            print(f"Indentation error in {file_path}, line {line_num}: "
                f"Expected {continued_indent} spaces, found {actual_indent}")
            return False
    else:
        if actual_indent != expected_indent:
            print(f"Indentation error in {file_path}, line {line_num}: "
                f"Expected {expected_indent} spaces, found {actual_indent}")
            return False
    return True

def check_indentation(file_path):
    procedure_indent = 2
    module_program_indent = 2
    loop_conditional_indent = 3
    continuation_indent = 5

    inside_module_program = False
    inside_procedure = False
    inside_loop_conditional = False
    inside_select = False
    specifier_line = False
    inside_derived_type = False
    on_continued_if_line = False
    inside_procedure_arguments = False
    inside_associate_arguments = False
    inside_do_concurrent_limits = False
    interface_block = False
    readwrite_line = False

    expected_indent = 0  # Default expected indentation
    continuation_line = False  # Flag to indicate if the previous line was a continuation
    open_bracket_count = 0  # Count of unbalanced open brackets
    close_bracket_count = 0  # Count of unbalanced close brackets
    unbalanced_brackets = 0
    continued_indent = 0
    num_single_quotes = 0
    num_double_quotes = 0
    unbalanced_quotes = False

    success = True

    with open(file_path, 'r') as file:
        for line_num, line in enumerate(file, start=1):
            
            # check length of line does not exceed
            if len(line) > 81:
                print(f"Line {line_num} in {file_path} exceeds 80 characters: {len(line)}")
                success = False

            stripped_line = line.rstrip()

            # Skip empty lines
            if not stripped_line:
                continuation_line = False
                continue

            # If comment line !###, skip
            if re.match(r'^\s*!###', stripped_line):
                continue

            # Check if preprocessing directive
            if re.match(r'^\s*#', stripped_line):
                continue

            # Check for lines with ! in first column and skip
            if re.match(r'^!', stripped_line):
                continue

            # If unbalanced quotes, check line starts with ampersand
            if unbalanced_quotes:
                if not re.match(r'^\s*&', stripped_line):
                    print(f"Unbalanced quotes in {file_path}, line {line_num}")
                    print(stripped_line)
                    print("Number of single quotes: ", num_single_quotes)
                    print("Number of double quotes: ", num_double_quotes)
                    return False

            # Replace all numbers at the start of the line with the same number of spaces
            stripped_line = re.sub(r'^\d+', lambda x: ' ' * len(x.group()), stripped_line)

            # Check if line starts with comment
            if re.match(r'^\s*!', stripped_line):
                actual_indent = len(stripped_line) - len(stripped_line.lstrip())
                if not check_if_match(actual_indent, expected_indent, continued_indent, continuation_line, line_num, file_path):
                    return False
                continue

            # Handle previous line with unbalanced quotes
            first_single_quote_pos = 0
            first_double_quote_pos = 0
            if num_single_quotes % 2 != 0:
                first_single_quote_pos = stripped_line.find("'") + 1
            if num_double_quotes % 2 != 0:
                first_double_quote_pos = stripped_line.find('"') + 1
            stripped_line_excld_quote = stripped_line[max(first_single_quote_pos, first_double_quote_pos):]

            # Check if line contains unbalanced quotes by counting the number of quotes
            if re.search(r"(?<!\\)'", stripped_line):
                num_single_quotes += stripped_line.count("'") - stripped_line.count("\\'")
            if re.search(r'(?<!\\)"', stripped_line):
                num_double_quotes += stripped_line.count('"') - stripped_line.count('\\"')

            # Check for quoted quotes and remove from the count, i.e. "'" or '"'
            quoted_quotes = re.findall(r'"[^"]*\'[^"]*"', stripped_line)
            if len(quoted_quotes) > 0:
                count = 0
                for q in quoted_quotes:
                    count += q.count(r'"')
                num_single_quotes -= count
            quoted_quotes = re.findall(r"'[^']*\"[^']*'", stripped_line)
            if len(quoted_quotes) > 0:
                count = 0
                for q in quoted_quotes:
                    count += q.count(r'"')
                num_double_quotes -= count

            # Handle unbalanced quotes
            last_single_quote_pos = len(stripped_line_excld_quote)
            last_double_quote_pos = len(stripped_line_excld_quote)
            if num_single_quotes % 2 != 0:
                last_single_quote_pos = stripped_line_excld_quote.rfind("'")
            if num_double_quotes % 2 != 0:
                last_double_quote_pos = stripped_line_excld_quote.rfind('"')
            stripped_line_excld_quote = stripped_line_excld_quote[:min(last_single_quote_pos, last_double_quote_pos)+1]
            
            # Check if line contains unbalanced quotes by counting the number of quotes
            if num_single_quotes % 2 != 0 or num_double_quotes % 2 != 0:
                unbalanced_quotes = True
            else:
                unbalanced_quotes = False
                num_single_quotes = 0
                num_double_quotes = 0

            # Check if line starts with close bracket, if so, update the indentation
            if re.match(r'^\s*(\)|/\)|\])', stripped_line_excld_quote):
                continued_indent = expected_indent + ( unbalanced_brackets - 1 ) * continuation_indent

            # Count open and close brackets
            open_bracket_count += stripped_line_excld_quote.count('(')
            open_bracket_count += stripped_line_excld_quote.count('[')
            close_bracket_count += stripped_line_excld_quote.count(')')
            close_bracket_count += stripped_line_excld_quote.count(']')
            if stripped_line_excld_quote.count('(') + stripped_line_excld_quote.count('[') < \
                    stripped_line_excld_quote.count(')') + stripped_line_excld_quote.count(']'):
                unbalanced_brackets -= 1
                if readwrite_line:
                    unbalanced_brackets += 1
            elif stripped_line_excld_quote.count('(') + stripped_line_excld_quote.count('[') > \
                    stripped_line_excld_quote.count(')') + stripped_line_excld_quote.count(']'):
                unbalanced_brackets += 1


            # Detect end of do loop, if statement, or where statement
            if re.match(r'^\s*end\s*(do|if|where|select)\b', stripped_line, re.IGNORECASE):
                expected_indent -= loop_conditional_indent

            # Detect else statements in if and where blocks, can be "PATTERN", "PATTERN\s*if", or "PATTERN\s*where"
            if inside_loop_conditional and re.match(r'^\s*else\s*(if|where)?\b', stripped_line, re.IGNORECASE):
                prior_indent = expected_indent
                expected_indent -= loop_conditional_indent
                specifier_line = True


            # Detect case, type, and rank statements within select, can be "PATTERN(", "PATTERN (" or "PATTERN default"
            if ( inside_select and re.match(r'^\s*(case|class is|type is|rank)\s*\(', stripped_line, re.IGNORECASE) ) or \
               ( inside_select and re.match(r'^\s*(case|class|rank)\s+default\b', stripped_line, re.IGNORECASE) ):
                prior_indent = expected_indent
                expected_indent -= loop_conditional_indent
                specifier_line = True

            # Detect if contains line
            if re.match(r'^\s*contains\b', stripped_line, re.IGNORECASE):
                prior_indent = expected_indent
                specifier_line = True
                if inside_derived_type:
                    expected_indent -= loop_conditional_indent - 1
                else:
                    expected_indent -= module_program_indent

            # Detect end of block block
            if re.match(r'^\s*end\s*block\b', stripped_line, re.IGNORECASE):
                expected_indent -= procedure_indent

            # Detect end of associate block
            if re.match(r'^\s*end\s*associate\b', stripped_line, re.IGNORECASE):
                expected_indent -= loop_conditional_indent

            # Detect end of interface block
            if re.match(r'^\s*end\s*interface\b', stripped_line, re.IGNORECASE):
                interface_block = False
                expected_indent -= loop_conditional_indent

            # Detect end of derived type block
            if inside_derived_type and re.match(r'^\s*end\s*type\b', stripped_line, re.IGNORECASE):
                expected_indent -= loop_conditional_indent
                inside_derived_type = False

            # Detect end of procedure block
            if inside_procedure and re.match(r'^\s*end\s*(function|subroutine|procedure)\b', stripped_line, re.IGNORECASE):
                expected_indent -= procedure_indent
                inside_procedure = False

            # Detect end of module or program
            if re.match(r'^\s*end\s*(module|program)\b', stripped_line, re.IGNORECASE):
                expected_indent -= module_program_indent



            # Check actual indentation
            actual_indent = len(stripped_line) - len(stripped_line.lstrip())
            if not check_if_match(actual_indent, expected_indent, continued_indent, continuation_line, line_num, file_path):
                return False
            

            # strip comments from end of line
            stripped_line = re.sub(r'!.*', '', stripped_line).strip()

            if continuation_line:
                continued_indent = expected_indent + unbalanced_brackets * continuation_indent

            # Check for line continuation character
            if stripped_line.endswith('&'):
                if not continuation_line:
                    continuation_line = True
                    # Set expected indentation for next line
                    continued_indent = expected_indent + continuation_indent
                    if unbalanced_brackets == 0:
                        unbalanced_brackets = 1
            else:
                # If it was a continuation line, reset to normal expected indentation
                if unbalanced_quotes:
                    print(f"Unbalanced quotes in {file_path}, line {line_num}")
                    return False
                open_bracket_count = 0
                close_bracket_count = 0
                unbalanced_brackets = 0
                if continuation_line:
                    continuation_line = False
                if inside_procedure_arguments:
                    inside_procedure_arguments = False
                    expected_indent += procedure_indent
                if inside_associate_arguments:
                    inside_associate_arguments = False
                    expected_indent += loop_conditional_indent
                if readwrite_line:
                    readwrite_line = False
                    

            # Reset from contains line
            if not continuation_line and specifier_line:
                specifier_line = False
                expected_indent = prior_indent
            

            # Detect module or program blocks (specifically avoid module procedure/function)
            if re.match(r'^\s*module\b(?!\s+(procedure|function))', stripped_line, re.IGNORECASE) or \
                re.match(r'^\s*program\b', stripped_line, re.IGNORECASE):
                expected_indent += module_program_indent

            # Detect procedure blocks, can be "module (function|subroutine|procedure)" or "(function|subroutine|procedure)" but not "procedure," or "procedure ::"
            if re.match(r'^\s*(integer\s+|logical\s+)?(pure\s+)?(module\s+)?(recursive\s+)?(function|subroutine|procedure)\b', stripped_line, re.IGNORECASE) and \
                not re.match(r'^\s*(function|subroutine|procedure)\s*(,|::)', stripped_line, re.IGNORECASE) and \
                not ( interface_block and re.match(r'^\s*procedure\s+\w+,', stripped_line, re.IGNORECASE) ):
                # print(stripped_line)
                # print(re.match(r'^\s*(integer\s+)', stripped_line, re.IGNORECASE), stripped_line)
                inside_procedure = True
                if stripped_line.lower().endswith("&"):
                    inside_procedure_arguments = True
                else:
                    expected_indent += procedure_indent


            # Detect derived type block
            if re.match(r'^\s*type\s*(::|,)', stripped_line, re.IGNORECASE):
                expected_indent += loop_conditional_indent
                inside_derived_type = True

            # Detect interface block, can be "abstract interface" or "interface"
            if re.match(r'^\s*(abstract\s+)?interface\b', stripped_line, re.IGNORECASE):
                interface_block = True
                expected_indent += loop_conditional_indent

            # Detect associate block
            if re.match(r'^\s*associate\b', stripped_line, re.IGNORECASE):
                if stripped_line.lower().endswith("&"):
                    inside_associate_arguments = True
                else:
                    expected_indent += loop_conditional_indent
            
            # Detect block block
            if re.match(r'^\s*block\b', stripped_line, re.IGNORECASE):
                expected_indent += procedure_indent

            # Detect do loop, and where statement with optional "NAME:"
            if re.match(r'^\s*\w+\s*:\s*(do|where)\b', stripped_line, re.IGNORECASE) or \
               re.match(r'^\s*(do|where)\b', stripped_line, re.IGNORECASE):
                expected_indent += loop_conditional_indent
                inside_loop_conditional = True

            # Detect do concurrent linebreak statement with optional "NAME:"
            if re.match(r'^\s*\w+\s*:\s*do\s+concurrent\b', stripped_line, re.IGNORECASE) or \
               re.match(r'^\s*do\s+concurrent\b', stripped_line, re.IGNORECASE):
                inside_do_concurrent_limits = True
                expected_indent -= loop_conditional_indent
                inside_loop_conditional = False
            
            if inside_do_concurrent_limits and not continuation_line:
                if stripped_line.lower().endswith(")"):
                    expected_indent += loop_conditional_indent
                inside_do_concurrent_limits = False
                inside_loop_conditional = True


            # Detect "if RANDOM then" statement with optional "NAME:"
            if re.match(r'^\s*\w*\s*:\s*if\s*\(.*\)\s*then\b', stripped_line, re.IGNORECASE) or \
               re.match(r'^\s*if\s*\(.*\)\s*then\b', stripped_line, re.IGNORECASE):
                expected_indent += loop_conditional_indent
                inside_loop_conditional = True

            # Detect line ends with "then" from unfinished if statement 
            if on_continued_if_line and not continuation_line:
                if stripped_line.lower().endswith("then"):
                    expected_indent += loop_conditional_indent
                on_continued_if_line = False

            # Detect if linebreak statement with optional "NAME:"
            if continuation_line and \
               ( re.match(r'^\s*\w+\s*:\s*(if)\b', stripped_line, re.IGNORECASE) or \
                 re.match(r'^\s*(if)\b', stripped_line, re.IGNORECASE) ):
                on_continued_if_line = True

            # Detect select type, select case, and select rank
            if re.match(r'^\s*select\s+(type|case|rank)\b', stripped_line, re.IGNORECASE):
                expected_indent += loop_conditional_indent
                inside_select = True

            # Detect read/write statement
            if re.match(r'^\s*(read|write)\(\b', stripped_line, re.IGNORECASE):
                readwrite_line = True


    return success


def main():
    # Check directory of script
    script_dir = os.path.dirname(os.path.abspath(__file__))
    os.chdir(script_dir+"/..")
    
    # Get list of .f90 and .F90 files from ../src/fortran and ../app
    fortran_files = glob.glob("src/fortran/**/*.f90", recursive=True) + glob.glob("src/fortran/**/*.F90", recursive=True)
    app_files = glob.glob("app/**/*.f90", recursive=True) + glob.glob("app/**/*.F90", recursive=True)
    all_files = fortran_files + app_files

    # Check indentation of all files
    success = True
    for file_path in all_files:
        if not check_indentation(file_path):
            success = False
        else:
            print(f"{file_path} passed indentation check.")

    # Exit with error if any file failed indentation check
    if not success:
        sys.exit("Indentation check failed.")
    else:
        print("All files passed indentation check.")

if __name__ == "__main__":
    main()



from mace.calculators import mace_mp
from ase.calculators.singlepoint import SinglePointCalculator
from raffle.generator import raffle_generator
from ase import build
from ase.optimize import FIRE
from ase.io import write
import numpy as np
import os
from joblib import Parallel, delayed

import logging
logging.basicConfig(level=logging.DEBUG)

# Function to relax a structure
def process_structure(i, atoms, num_structures_old, calc_params, optimise_structure, iteration, calc):
    # Check if the structure has already been processed
    if i < num_structures_old:
        return None, None, None
    
    # calc = Vasp(**calc_params, label=f"struct{i}", directory=f"iteration{iteration}/struct{i}/", txt=f"stdout{i}.o")
    inew = i - num_structures_old
    atoms.calc = calc

    # Calculate and save the initial energy per atom
    energy_unrlxd = atoms.get_potential_energy() / len(atoms)
    print(f"Initial energy per atom: {energy_unrlxd}")

    # Optimise the structure if requested
    if optimise_structure:
        optimizer = FIRE(atoms, trajectory = f"traje{inew}.traj", logfile=f"optimisation{inew}.log")
        try:
            optimizer.run(fmax=0.05, steps=100)
        except Exception as e:
            print(f"Optimisation failed: {e}")
            return None, None, None
    
    # Save the optimised structure and its energy per atom
    energy_rlxd = atoms.get_potential_energy() / len(atoms)

    # Get the distance matrix
    distances = atoms.get_all_distances(mic=True)

    # Set self-distances (diagonal entries) to infinity to ignore them
    np.fill_diagonal(distances, np.inf)

    # Check if the minimum non-self distance is below 1.5
    if distances.min() < 1.0:
        print(f"Distance too small: {atoms.get_all_distances(mic=True).min()}")
        return None, None, None
    
    if abs(energy_rlxd - energy_unrlxd) > 10.0:
        print(f"Energy difference too large: {energy_rlxd} vs {energy_unrlxd}")
        return None, None, None
    
    return atoms, energy_unrlxd, energy_rlxd


if __name__ == "__main__":

    # set up the calculator
    calc_params = { 'model': "../mace-mpa-0-medium.model" }
    calc = mace_mp(**calc_params)

    # set up the hosts
    hosts = []

    Si_bulk = build.bulk("Si", crystalstructure="diamond", a=5.43)
    Si_bulk.calc = calc
    Si_reference_energy = Si_bulk.get_potential_energy() / len(Si_bulk)
    Si_cubic = build.make_supercell(Si_bulk, [[-1, 1, 1], [1, -1, 1], [1, 1, -1]])
    Ge_bulk = build.bulk("Ge", crystalstructure="diamond", a=5.65)
    Ge_bulk.calc = calc
    Ge_cubic = build.make_supercell(Ge_bulk, [[-1, 1, 1], [1, -1, 1], [1, 1, -1]])
    Ge_reference_energy = Ge_bulk.get_potential_energy() / len(Ge_bulk)

    Si_supercell = build.make_supercell(Si_cubic, [[2, 0, 0], [0, 2, 0], [0, 0, 1]])
    Ge_supercell = build.make_supercell(Ge_cubic, [[2, 0, 0], [0, 2, 0], [0, 0, 1]])

    Si_surface = build.surface(Si_supercell, indices=(0, 0, 1), layers=2)
    Ge_surface = build.surface(Ge_supercell, indices=(0, 0, 1), layers=2)

    Si_slab = build.surface(Si_supercell, indices=(0, 0, 1), layers=2, vacuum=12, periodic=True)
    Si_slab.calc = calc
    Ge_slab = build.surface(Ge_supercell, indices=(0, 0, 1), layers=2, vacuum=12, periodic=True)
    Ge_slab.calc = calc

    host = build.stack(Si_surface, Ge_surface, axis=2, distance= 5.43/2 + 5.65/2)
    cell = host.get_cell()
    print("Host cell: ", host.get_cell())
    cell[2, 2] -= 3.8865 # (5.43 + 5.65) / 2 * 3/4
    host.set_cell(cell, scale_atoms=False)

    hosts.append(host)

    # set the parameters for the generator
    generator = raffle_generator()
    generator.distributions.set_element_energies(
        {
            'Si': Si_reference_energy,
            'Ge': Ge_reference_energy,
        }
    )

    # set energy scale
    generator.distributions.set_kBT(0.2)

    # set the distribution function widths (2-body, 3-body, 4-body)
    generator.distributions.set_width([0.04, np.pi/160.0, np.pi/160.0])

    # set the initial database
    initial_database = [Si_bulk, Ge_bulk]
    generator.distributions.create(initial_database, deallocate_systems=False)

    # set the parameters for the generator
    seed = 0

    # check if the energies file exists, if not create it
    energies_rlxd_filename = f"energies_rlxd_seed{seed}.txt"
    energies_unrlxd_filename = f"energies_unrlxd_seed{seed}.txt"
    
    if os.path.exists(energies_rlxd_filename):
        with open(energies_rlxd_filename, "w") as energy_file:
            pass
    else:
        open(energies_rlxd_filename, "w").close()

    if os.path.exists(energies_unrlxd_filename):
        with open(energies_unrlxd_filename, "w") as energy_file:
            pass
    else:
        open(energies_unrlxd_filename, "w").close()
        
    # initialise the number of structures generated
    iter = -1
    unrlxd_structures = []
    rlxd_structures = []
    num_structures_old = 0
    optimise_structure = True
    # start the iterations, loop over the hosts, then repeat the process X times
    for ival in range(40):
        for host in hosts:
            print("setting host")
            generator.set_host(host)
            generator.set_bounds([[0, 0, 0.34], [1, 1, 0.52]])

            iter += 1
            print(f"Iteration {iter}")

            # generate the structures
            generator.generate(
                num_structures = 5,
                stoichiometry = { 'Si': 16, 'Ge': 16 },
                seed = seed*1000+iter,
                method_ratio = {"void": 0.1, "rand": 0.01, "walk": 0.25, "grow": 0.25, "min": 1.0},
                verbose = 0,
            )

            # print the number of structures generated
            print("Total number of structures generated: ", generator.num_structures)
            generated_structures = generator.get_structures(calc)
            num_structures_new = len(generated_structures)

            # check if directory iteration[iter] exists, if not create it
            iterdir = f"iteration{iter}/"
            if not os.path.exists(iterdir):
                os.makedirs(iterdir)
            generator.print_settings(iterdir+"generator_settings.txt")

            # set up list of energies
            energy_unrlxd = np.zeros(num_structures_new - num_structures_old)
            energy_rlxd = np.zeros(num_structures_new - num_structures_old)
            for i in range(num_structures_new - num_structures_old):
                write(iterdir+f"POSCAR_unrlxd_{i}", generated_structures[num_structures_old + i])
                print(f"Structure {i} energy per atom: {generated_structures[num_structures_old + i].get_potential_energy() / len(generated_structures[num_structures_old + i])}")
                # print(f"Structure {i} energy per unit area: {( generated_structures[num_structures_old + i].get_potential_energy() - Si_slab.get_potential_energy() - Ge_slab.get_potential_energy() ) / (2 * ( cell[0, 0] * cell[1, 1] ) )}")
                unrlxd_structures.append(generated_structures[num_structures_old + i].copy())
                unrlxd_structures[-1].calc = SinglePointCalculator(
                    generated_structures[num_structures_old + i],
                    energy=generated_structures[num_structures_old + i].get_potential_energy(),
                    forces=generated_structures[num_structures_old + i].get_forces()
                )
            
            # Start parallel execution
            print("Starting parallel execution")
            results = Parallel(n_jobs=5)(
                delayed(process_structure)(i, generated_structures[i].copy(), num_structures_old, calc_params, optimise_structure, iteration=seed, calc=calc)
                for i in range(num_structures_old, num_structures_new)
            )

            # Wait for all futures to complete
            for j, result in enumerate(results):
                generated_structures[num_structures_old + j], energy_unrlxd[j], energy_rlxd[j] = result
                if generated_structures[num_structures_old + j] is None:
                    print("Structure failed the checks")
                    continue
                rlxd_structures.append(generated_structures[num_structures_old + j].copy())
                rlxd_structures[-1].calc = SinglePointCalculator(
                    generated_structures[j+num_structures_old],
                    energy=generated_structures[num_structures_old + j].get_potential_energy(),
                    forces=generated_structures[num_structures_old + j].get_forces()
                )
            print("All futures completed")

            # Remove structures that failed the checks
            for j, atoms in reversed(list(enumerate(generated_structures))):
                if j < num_structures_old:
                    continue
                if atoms is None:
                    energy_unrlxd = np.delete(energy_unrlxd, j-num_structures_old)
                    energy_rlxd = np.delete(energy_rlxd, j-num_structures_old)
                    del generated_structures[j]
                    del unrlxd_structures[j]
                    del rlxd_structures[j]
                    generator.remove_structure(j)
            num_structures_new = len(generated_structures) 

            # write the structures to files
            for i in range(num_structures_new - num_structures_old):
                write(iterdir+f"POSCAR_{i}", generated_structures[num_structures_old + i])
                print(f"Structure {i} energy per atom: {energy_rlxd[i]}")
                # append energy per atom to the 'energies_unrlxd_filename' file
                with open(energies_unrlxd_filename, "a") as energy_file:
                    energy_file.write(f"{i+num_structures_old} {energy_unrlxd[i]}\n")
                # append energy per atom to the 'energies_rlxd_filename' file
                with open(energies_rlxd_filename, "a") as energy_file:
                    energy_file.write(f"{i+num_structures_old} {energy_rlxd[i]}\n")

            # update the distribution functions
            print("Updating distributions")
            generator.distributions.update(generated_structures[num_structures_old:], from_host=False, deallocate_systems=False)

            # print the new distribution functions to a file
            print("Printing distributions")
            generator.distributions.write_dfs(iterdir+"distributions.txt")
            generator.distributions.write_2body(iterdir+"df2.txt")
            generator.distributions.write_3body(iterdir+"df3.txt")
            generator.distributions.write_4body(iterdir+"df4.txt")
            generator.distributions.deallocate_systems()

            # update the number of structures generated
            num_structures_old = num_structures_new


    # Write the final distribution functions to a file
    generator.distributions.write_gdfs(f"gdfs_seed{seed}.txt")

    # Read energies from the file
    with open(energies_rlxd_filename, "r") as energy_file:
        energies = energy_file.readlines()

    # Parse and sort the energies
    energies = [line.strip().split() for line in energies]
    energies = sorted(energies, key=lambda x: float(x[1]))

    # Write the sorted energies back to the file
    with open(f"sorted_{energies_rlxd_filename}", "w") as energy_file:
        for entry in energies:
            energy_file.write(f"{int(entry[0])} {float(entry[1])}\n")

    # Write the structures to files
    write(f"unrlxd_structures_seed{seed}.traj", unrlxd_structures)
    write(f"rlxd_structures_seed{seed}.traj", rlxd_structures)
    print("All generated and relaxed structures written")

    print("Learning complete")


import os
import numpy as np

## import ASE (Atomic Simulation Environment) modules
from ase import Atoms
from ase.io import read, write


## load calculator
calculator = "CHGNet"
match calculator:
    case "CHGNet":
        from chgnet.model.dynamics import CHGNetCalculator
        print("Initialising CHGNet calculator")
        calc = CHGNetCalculator()
        label = "CHGNet"
    case "MACE":
        from mace.calculators import mace_mp
        print("Initialising MACE calculator")
        calc = mace_mp(model="medium", dispersion=False, default_dtype="float32", device='cpu')
        label = "MACE"

## Read the database
print("Reading database")
database = read("../example/data/Si-Ge.xyz", index=":")

Si_energy_dft = -10.644538 / 2 # Si POT
Ge_energy_dft = -8.7301811 / 2 # Ge POT

## Calculate the energies
energies_dft = []
energies_mlp = []
formation_energies_dft = []
formation_energies_mlp = []
for i, atoms in enumerate(database):
    if atoms.calc is None:
        database.remove(atoms)
        continue
    # get number of Si atoms in the structure
    n_Si = atoms.get_atomic_numbers().tolist().count(14)
    n_Ge = atoms.get_atomic_numbers().tolist().count(32)
    if(n_Si == 0 and n_Ge == 0 or n_Si + n_Ge != len(atoms)):
        print(f"Skipping structure {i} as it has no Si or Ge atoms")
        continue
    # if(n_Si == 0 or n_Ge == 0):
    #     print(f"Skipping structure {i} as it has only one type of atom")
    #     continue
    formation_energy_dft = ( atoms.get_potential_energy() - n_Si * Si_energy_dft - n_Ge * Ge_energy_dft ) / len(atoms)
    formation_energies_dft.append(formation_energy_dft)
    energies_dft.append(atoms.get_potential_energy()/len(atoms))
    atoms.calc = mace_calc
    formation_energy_mace = ( atoms.get_potential_energy() - n_Si * Si_energy_dft - n_Ge * Ge_energy_dft ) / len(atoms)
    formation_energies_mlp.append(formation_energy_mace)
    energies_mlp.append(atoms.get_potential_energy()/len(atoms))
    print(n_Si, n_Ge, formation_energy_dft, formation_energy_mace, atoms.get_potential_energy()/len(atoms))
    # if energies_mlp[-1] - energies_dft[-1] > 3e-1 and energies_mlp[-1] < -7.9:
    #     print(f"Energy difference for structure {i} is {energies_mlp[-1] - energies_dft[-1]}, energy_mace: {energies_mlp[-1]}")
    #     view(atoms)


import matplotlib.pyplot as plt

## Write energies to a file
with open("Si-Ge_energies_comparison.txt", "w") as f:
    f.write("# DFT_Energy_per_atom MACE_Energy_per_atom\n")
    for dft_energy, mace_energy in zip(energies_dft, energies_mlp):
        f.write(f"{dft_energy} {mace_energy}\n")

with open("Si-Ge_formations_comparison.txt", "w") as f:
    f.write("# DFT_Formation_energy_per_atom "+label+"_Formation_energy_per_atom\n")
    for dft_energy, mace_energy in zip(formation_energies_dft, formation_energies_mlp):
        f.write(f"{dft_energy} {mace_energy}\n")

## Plotting the energies
plt.figure(figsize=(10, 6))
plt.scatter(energies_dft, energies_mlp, c='blue', marker='o', label=label+' vs DFT')
plt.scatter(formation_energies_dft, formation_energies_mlp, c='blue', marker='o', label=label+' vs DFT')
plt.show()